from tomwer.tests.conftest import qtapp  # noqa F401
import asyncio
import os

import pytest

from orangecontrib.tomwer.widgets.cluster.FutureSupervisorOW import FutureSupervisorOW
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
def test_FutureSupervisorOW(
    qtapp,  # noqa F811
    tmp_path,
):
    window = FutureSupervisorOW()
    test_dir = tmp_path / "scans_folder"
    test_dir.mkdir()

    # set up scans
    scans = []
    future_tomo_objs = []
    for i in range(5):
        # create scan
        scan = MockNXtomo(
            scan_path=os.path.join(test_dir, f"scan_test{i}"),
            n_proj=10,
            n_ini_proj=10,
            create_ini_dark=False,
            create_ini_flat=False,
            dim=10,
        ).scan
        scans.append(scan)

        # create future
        future = asyncio.Future()
        if i == 1:
            future.set_result(None)
        future_tomo_objs.append(
            FutureTomwerObject(
                tomo_obj=scan,
                futures=(future,),
            )
        )

    window.close()



import os

import numpy
import pint
from nxtomo.application.nxtomo import NXtomo
from silx.gui import qt
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.visualization.nxtomometadata import NXtomoMetadataViewer
from tomwer.tests.conftest import qtapp  # noqa F401

_ureg = pint.get_application_registry()


def test_nx_editor(
    tmp_path,
    qtapp,  # noqa F811
):
    # 1.0 create nx tomo with raw data
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.x_pixel_size = 2.6e-6 * _ureg.meter
    nx_tomo.instrument.detector.y_pixel_size = 2.5e-6 * _ureg.meter
    nx_tomo.instrument.detector.field_of_view = "Half"
    nx_tomo.instrument.detector.distance = 59.0 * _ureg.meter
    nx_tomo.instrument.detector.set_transformation_from_x_flipped(True)
    nx_tomo.instrument.detector.set_transformation_from_y_flipped(False)
    nx_tomo.energy = 12.8 * _ureg.keV
    # warning: NXtomo uses McStas coordinate system
    nx_tomo.sample.z_translation = numpy.arange(12) * _ureg.meter
    nx_tomo.sample.y_translation = numpy.arange(2, 14) * _ureg.meter
    nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo.sample.rotation_angle = numpy.linspace(0, 180, num=12) * _ureg.degree

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo.save(
        file_path=file_path,
        data_path=entry,
    )

    scan = NXtomoScan(file_path, entry)

    # 2.0 create the widget and do the edition
    widget = NXtomoMetadataViewer()
    widget.setScan(scan=scan)
    widget.show()

    # 3.0 check data have been corrcetly loaded
    def check_metric(expected_value, current_value):
        if expected_value is None:
            return current_value is None
        return expected_value == current_value

    assert check_metric(
        2.6e-6,
        widget._xDetectorPixelSizeMetricEntry.getValue().to(_ureg.meter).magnitude,
    )
    assert widget._xDetectorPixelSizeMetricEntry._qcbUnit.currentText() == "m"
    assert check_metric(
        2.5e-6,
        widget._yDetectorPixelSizeMetricEntry.getValue().to(_ureg.meter).magnitude,
    )
    assert widget._yDetectorPixelSizeMetricEntry._qcbUnit.currentText() == "m"

    assert check_metric(
        59,
        widget._sampleDetectorDistanceMetricEntry.getValue().to(_ureg.meter).magnitude,
    )
    assert widget._sampleDetectorDistanceMetricEntry._qcbUnit.currentText() == "m"

    assert "Half" == widget._fieldOfViewCB.currentText()
    assert widget._lrFlippedCB.isChecked()
    assert not widget._udFlippedCB.isChecked()

    assert "12.8" == widget._energyEntry.text()

    # end
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)
    widget.close()
    widget = None



# coding: utf-8
from __future__ import annotations


import pytest
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.gui.visualization.reconstructionparameters import ReconstructionParameters

from tomwer.tests.conftest import qtapp  # noqa F401


@pytest.mark.parametrize("phase_method", ("", "CTF", "Paganin"))
def test_ReconstructionParameters(qtapp, phase_method):  # noqa F401
    window = ReconstructionParameters()
    volume = HDF5Volume(
        file_path="test.hdf5",
        data_path="data",
        data=None,
        metadata={
            "nabu_config": {
                "reconstruction": {
                    "method": "FBP",
                },
                "phase": {
                    "method": phase_method,
                    "delta_beta": 110.0,
                },
            },
            "processing_options": {
                "reconstruction": {
                    "voxel_size_cm": (0.2, 0.2, 0.2),
                    "rotation_axis_position": 104,
                    "enable_halftomo": True,
                    "fbp_filter_type": "Hilbert",
                    "sample_detector_dist": 0.4,
                },
                "take_log": {
                    "log_min_clip": 1.0,
                    "log_max_clip": 10.0,
                },
            },
        },
    )
    window.setVolumeMetadata(metadata=volume.metadata)

    assert window._methodQLE.text() == "FBP"
    assert window._paganinQLE.text() == phase_method
    assert window._deltaBetaQLE.text() == "110.0"
    assert window._sampleDetectorDistanceInMeterQLE.text() == "0.004"
    assert window._voxelSizeInMicronQLE.text() == "2e+03x2e+03x2e+03"
    assert window._corQLE.text() == "104.00"
    assert window._halfTomoCB.isChecked()
    assert window._fbpFilterQLE.text() == "Hilbert"
    assert window._minLogClipQLE.text() == "1.0"
    assert window._maxLogClipQLE.text() == "10.0"

    window.setVolumeMetadata(metadata=None)



# coding: utf-8
from __future__ import annotations

import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.utils import inputwidget
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class SelectionLineEditTest(TestCaseQt):
    """Test the SelectionLineEdit"""

    def setUp(self):
        super().setUp()
        self.widget = inputwidget.SelectionLineEdit(parent=None)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        super().tearDown()

    def testListSelection(self):
        self.widget.mode = inputwidget.SelectionLineEdit.LIST_MODE
        self.widget.setText("1.0, 2.0; 6.3")
        self.assertTrue(self.widget.selection == (1.0, 2.0, 6.3))
        self.widget.setText("1.0:3.6:0.2")
        self.assertTrue(
            self.widget.selection == tuple(numpy.linspace(1.0, 3.6, num=int(2.6 / 0.2)))
        )
        self.assertTrue(
            self.widget.getMode() == inputwidget.SelectionLineEdit.RANGE_MODE
        )
        self.widget.setText("1.0")
        self.assertTrue(self.widget.selection == 1.0)

    def testRangeSelection(self):
        self.widget.mode = inputwidget.SelectionLineEdit.RANGE_MODE
        self.widget.setText("1.0:3.6:0.2")
        self.assertTrue(
            self.widget.selection == tuple(numpy.linspace(1.0, 3.6, num=int(2.6 / 0.2)))
        )
        self.widget.setText("1.0")
        self.assertTrue(self.widget.selection == 1.0)
        self.widget.setText("1.0, 2.0, 6.3")
        self.assertTrue(
            self.widget.getMode() == inputwidget.SelectionLineEdit.LIST_MODE
        )
        self.assertTrue(self.widget.selection == (1.0, 2.0, 6.3))



"""contains QLFileSystem class. Specific implementation of the QLineEdit to select a file path."""

from __future__ import annotations


from silx.gui import qt


class QLFileSystem(qt.QLineEdit):
    """
    QLineEdit with a completer using a QDirModel
    """

    def __init__(self, text, parent, filters=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.completer = create_completer(filters=filters)
        self.setCompleter(self.completer)
        if text is not None:
            self.setText(text)


def create_completer(filters=None) -> qt.QCompleter:
    completer = qt.QCompleter()
    if qt.BINDING in ("PyQt5", "PySide2"):
        # note: should work with PyQt5 but we get some troubles with it on esrf deployment
        # see https://gitlab.esrf.fr/XRD/darfix/-/issues/174
        model = qt.QDirModel(completer)
    else:
        completer.setCompletionRole(qt.QFileSystemModel.FilePathRole)
        model = qt.QFileSystemModel(completer)
        model.setRootPath(qt.QDir.currentPath())
        model.setOption(qt.QFileSystemModel.DontWatchForChanges)
    if filters is not None:
        model.setFilter(filters)
    completer.setModel(model)

    return completer



from __future__ import annotations


from silx.gui import qt
from silx.gui.plot import actions
from tomwer.gui import icons as tomwer_icons

from .actions._ColorBarAction import ColorBarAction
from .actions._ShowAxesAction import ShowAxesAction
from .actions._ShowCoordinateSystemAction import ShowCoordinateSystemAction
from .actions._ShowCoordinateSystemLabelAction import ShowCoordinateSystemLabelAction
from .actions._ShowPlotToolBarAction import ShowPlotToolBarAction


class ToolBar(qt.QToolBar):
    """Specialized toolbar for the VolumeReconstructionSummaryWindow"""

    def __init__(self, master_plot, colorbar, plots, parent=None):
        super().__init__(parent=parent)

        # colormap action
        self._colormapAction = actions.control.ColormapAction(
            parent=self, plot=master_plot
        )
        master_plot.sigActiveImageChanged.disconnect(
            self._colormapAction._updateColormap
        )
        master_plot.sigActiveScatterChanged.disconnect(
            self._colormapAction._updateColormap
        )
        self.addAction(self._colormapAction)

        # colorbar action
        self.colorbarAction = ColorBarAction(colorbar=colorbar, parent=self)
        self.colorbarAction.setVisible(True)
        self.colorbarAction.setChecked(True)
        self.addAction(self.colorbarAction)

        # show axes
        self.showAxesAction = ShowAxesAction(parent=self, plots=plots)
        self.addAction(self.showAxesAction)

        # Toggle coordinate system
        self.showCoordinateSystem = ShowCoordinateSystemAction(
            plots=[plot.parent() for plot in plots], parent=self
        )
        self.addAction(self.showCoordinateSystem)

        # Toggle labels
        self.showCoordinatesSystemLabel = ShowCoordinateSystemLabelAction(
            plots=[plot.parent() for plot in plots], parent=self
        )
        self.addAction(self.showCoordinatesSystemLabel)

        # Toggle toolbar action
        self.showToolbar = ShowPlotToolBarAction(
            plots=[plot.parent() for plot in plots], parent=self
        )
        self.addAction(self.showToolbar)

        # separator
        self.addSeparator()

        # load volume
        self._loadVolumeWidget = qt.QCheckBox("load volume")
        self._loadVolumeWidget.setToolTip(
            "Loading the volume in memory will allow you to browse the slices along the three dimensions. But it will consume a lot of RAM."
        )
        self._loadVolumeAction = self.addWidget(self._loadVolumeWidget)

        # separator
        self.addSeparator()

        # display a single axis
        axis0_icon = tomwer_icons.getQIcon("extend_3D_coordinate_system_XY")
        self._extendXYPlotAction = qt.QAction(axis0_icon, "extend XY plot")
        self.addAction(self._extendXYPlotAction)

        axis1_icon = tomwer_icons.getQIcon("extend_3D_coordinate_system_XZ")
        self._extendXZPlotAction = qt.QAction(axis1_icon, "extend XZ plot")
        self.addAction(self._extendXZPlotAction)

        axis2_icon = tomwer_icons.getQIcon("extend_3D_coordinate_system_YZ")
        self._extendYZPlotAction = qt.QAction(axis2_icon, "extend YZ plot")
        self.addAction(self._extendYZPlotAction)

        # equally space plots
        reset_display_icon = tomwer_icons.getQIcon("reset_3D_coordinate_system")
        self._resetDisplayAction = qt.QAction(
            reset_display_icon, "Reset plots position."
        )
        self.addAction(self._resetDisplayAction)

    # Action getters
    def getLoadVolumeAction(self):
        return self._loadVolumeAction

    def getExtendXYPlotAction(self) -> qt.QAction:
        return self._extendXYPlotAction

    def getExtendXZPlotAction(self) -> qt.QAction:
        return self._extendXZPlotAction

    def getExtendYZPlotAction(self) -> qt.QAction:
        return self._extendYZPlotAction

    def getResetDisplayAction(self) -> qt.QAction:
        return self._resetDisplayAction



# coding: utf-8

from ewokscore.task import Task as EwoksTask


class _SinogramViewerPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to display a sinogram
    """

    def run(self):
        pass



from __future__ import annotations

import os
import h5py

import numpy
import pytest
import pint
from silx.gui import qt

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey, FOV
from nxtomo.utils.transformation import (
    build_matrix,
    DetXFlipTransformation,
    DetYFlipTransformation,
)

from tomwer.core.process.edit.nxtomoeditor import NXtomoEditorTask
from nxtomo.nxobject.nxtransformations import (
    NXtransformations,
    get_lr_flip,
    get_ud_flip,
)
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.edit.nxtomoeditor import NXtomoEditor, _TranslationMetricEntry
from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.tests.test_utils import is_lr_flip, is_ud_flip

_ureg = pint.get_application_registry()


@pytest.mark.parametrize("detector_x_pixel_size", (0.12 * _ureg.millimeter,))
@pytest.mark.parametrize("detector_y_pixel_size", (None,))
@pytest.mark.parametrize("sample_x_pixel_size", (None,))
@pytest.mark.parametrize("sample_y_pixel_size", (None, 0.0066 * _ureg.meter))
@pytest.mark.parametrize("field_of_view", [item.value for item in FOV])
@pytest.mark.parametrize("sample_detector_distance", (1.2 * _ureg.meter,))
@pytest.mark.parametrize("sample_source_distance", (-10.2 * _ureg.meter,))
@pytest.mark.parametrize("propagation_distance", (1.01 * _ureg.meter,))
@pytest.mark.parametrize("energy", (None, 23.5 * _ureg.keV))
@pytest.mark.parametrize("lr_flipped", (False,))
@pytest.mark.parametrize("ud_flipped", (True, False))
@pytest.mark.parametrize(
    "x_translation",
    (None, numpy.ones(12) * _ureg.meter, numpy.arange(12) * _ureg.meter),
)
@pytest.mark.parametrize(
    "z_translation",
    (None, numpy.zeros(12) * _ureg.meter, numpy.arange(12, 24) * _ureg.meter),
)
def test_nx_editor(
    tmp_path,
    qtapp,  # noqa F811
    detector_x_pixel_size: pint.Quantity | None,
    detector_y_pixel_size: pint.Quantity | None,
    sample_x_pixel_size: pint.Quantity | None,
    sample_y_pixel_size: pint.Quantity | None,
    propagation_distance: pint.Quantity | None,
    sample_source_distance: pint.Quantity | None,
    field_of_view,
    sample_detector_distance: pint.Quantity,
    energy: pint.Quantity,
    lr_flipped,
    ud_flipped,
    x_translation: pint.Quantity,
    z_translation: pint.Quantity,
):
    """
    :param x_translation: x translation on the ESRF coordinate system (will be converted to McStas coordinate when saved)
    :param z_translation: z translation on the ESRF coordinate system (will be converted to McStas coordinate when saved)
    """
    # 1.0 create nx tomo with raw data
    nx_tomo = NXtomo()
    nx_tomo.energy = energy
    nx_tomo.instrument.detector.x_pixel_size = detector_x_pixel_size
    nx_tomo.instrument.detector.y_pixel_size = detector_y_pixel_size
    nx_tomo.instrument.detector.field_of_view = field_of_view
    nx_tomo.instrument.detector.distance = sample_detector_distance
    nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))

    nx_tomo.sample.rotation_angle = numpy.linspace(0, 20, num=12) * _ureg.degree
    # warning: mapping from ESRF coordinate to McStas
    nx_tomo.sample.z_translation = x_translation
    nx_tomo.sample.y_translation = z_translation
    nx_tomo.sample.x_pixel_size = sample_x_pixel_size
    nx_tomo.sample.y_pixel_size = sample_y_pixel_size
    nx_tomo.sample.propagation_distance = propagation_distance

    nx_tomo.instrument.source.distance = sample_source_distance

    nx_tomo.instrument.detector.set_transformation_from_lr_flipped(flipped=lr_flipped)
    nx_tomo.instrument.detector.set_transformation_from_ud_flipped(flipped=ud_flipped)

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo.save(
        file_path=file_path,
        data_path=entry,
    )

    scan = NXtomoScan(file_path, entry)

    # 2.0 create the widget and do the edition
    widget = NXtomoEditor()
    widget.setScan(scan=scan)

    # 3.0 check data have been correctly loaded
    def check_metric(expected_value, current_value):
        if expected_value is None:
            return current_value is None
        return numpy.isclose(expected_value, current_value)

    assert check_metric(
        detector_x_pixel_size, widget._xDetectorPixelSizeMetricEntry.getValue()
    )
    assert widget._xDetectorPixelSizeMetricEntry._qcbUnit.currentText() == "m"
    assert check_metric(
        detector_y_pixel_size, widget._yDetectorPixelSizeMetricEntry.getValue()
    )
    assert widget._yDetectorPixelSizeMetricEntry._qcbUnit.currentText() == "m"
    assert check_metric(
        sample_detector_distance, widget._sampleDetectorDistanceMetricEntry.getValue()
    )
    assert widget._sampleDetectorDistanceMetricEntry._qcbUnit.currentText() == "m"
    assert check_metric(
        sample_x_pixel_size, widget._xSamplePixelSizeMetricEntry.getValue()
    )
    assert check_metric(
        sample_y_pixel_size, widget._ySamplePixelSizeMetricEntry.getValue()
    )
    assert check_metric(
        sample_source_distance, widget._sampleSourceDistanceMetricEntry.getValue()
    )
    assert check_metric(
        propagation_distance, widget._propagationDistanceMetricEntry.getValue()
    )
    assert field_of_view == widget._fieldOfViewCB.currentText()
    assert lr_flipped == widget._lrFlippedCB.isChecked()
    assert ud_flipped == widget._udFlippedCB.isChecked()

    if energy is None:
        assert widget._energyEntry.getValue() is None
    else:
        assert numpy.isclose(
            energy.to(_ureg.keV).magnitude, widget._energyEntry.getValue()
        )

    def check_translation(expected_value, current_value):
        if expected_value is None:
            return current_value is None
        else:
            expected_value = expected_value.to_base_units().magnitude
            assert current_value is not None
            if isinstance(current_value, pint.Quantity):
                current_value = current_value.to_base_units().magnitude
            u_values = numpy.unique(expected_value)
            if u_values.size == 1:
                return float(current_value) == u_values[0]
            else:
                return current_value is _TranslationMetricEntry.LOADED_ARRAY

    assert check_translation(x_translation, widget._xTranslationQLE.getValue())
    assert widget._xTranslationQLE._qcbUnit.currentText() == "m"
    assert check_translation(z_translation, widget._zTranslationQLE.getValue())
    assert widget._zTranslationQLE._qcbUnit.currentText() == "m"

    # 4.0 edit some parameters
    widget._energyEntry.setText("23.789")
    widget._xDetectorPixelSizeMetricEntry.setUnit(_ureg.nanometer)
    widget._yDetectorPixelSizeMetricEntry.setValue(2.1e-7)
    widget._xSamplePixelSizeMetricEntry.setValue(value_m=5.6e-6)
    widget._sampleDetectorDistanceMetricEntry.setValue("unknown")
    widget._sampleSourceDistanceMetricEntry.setValue(-99)
    widget._propagationDistanceMetricEntry.setValue(88)
    widget._fieldOfViewCB.setCurrentText(FOV.HALF.value)
    widget._lrFlippedCB.setChecked(not lr_flipped)
    widget._xTranslationQLE.setValue(1.8)
    widget._xTranslationQLE.setUnit(_ureg.millimeter)
    widget._zTranslationQLE.setValue(2.8)
    widget._zTranslationQLE.setUnit(_ureg.meter)

    # 5.0
    task = NXtomoEditorTask(
        inputs={
            "data": scan,
            "configuration": widget.getConfigurationForTask(),
        }
    )
    task.run()

    # 6.0 make sure data have been overwrite
    overwrite_nx_tomo = NXtomo().load(
        file_path=file_path,
        data_path=entry,
    )

    assert overwrite_nx_tomo.energy == 23.789 * _ureg.keV

    if detector_x_pixel_size is None:
        assert overwrite_nx_tomo.instrument.detector.x_pixel_size is None
    else:
        assert numpy.isclose(
            overwrite_nx_tomo.instrument.detector.x_pixel_size,
            0.12 * _ureg.nanometer,
        )
    assert overwrite_nx_tomo.instrument.detector.y_pixel_size == 2.1e-7 * _ureg.meter

    assert overwrite_nx_tomo.instrument.detector.distance is None
    assert overwrite_nx_tomo.instrument.detector.field_of_view is FOV.HALF

    final_transformation = NXtransformations()

    final_transformation.add_transformation(DetXFlipTransformation(flip=ud_flipped))
    final_transformation.add_transformation(DetYFlipTransformation(flip=not lr_flipped))
    # note: the 'not' comes from inversion done with the _xFlippedCB combobox

    numpy.testing.assert_allclose(
        build_matrix(
            overwrite_nx_tomo.instrument.detector.transformations.transformations
        ),
        build_matrix(final_transformation.transformations),
    )

    numpy.testing.assert_array_almost_equal(
        overwrite_nx_tomo.sample.x_translation,
        (numpy.array([1.8] * 12) * _ureg.millimeter).to_base_units(),
    )
    numpy.testing.assert_array_almost_equal(
        overwrite_nx_tomo.sample.z_translation,
        numpy.array([2.8] * 12) * _ureg.meter,
    )
    # end
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)
    widget.close()
    widget = None


def test_nx_editor_lock(
    tmp_path,
    qtapp,  # noqa F811
):
    """test the pad lock buttons of the NXtomo editor"""
    # 1.0 create nx tomos with raw data
    nx_tomo_1 = NXtomo()
    nx_tomo_1.instrument.source.distance = 54.8 * _ureg.meter
    nx_tomo_1.instrument.detector.x_pixel_size = 0.023 * _ureg.meter
    nx_tomo_1.instrument.detector.y_pixel_size = 0.025 * _ureg.meter
    nx_tomo_1.instrument.detector.field_of_view = "full"
    nx_tomo_1.instrument.detector.distance = 2.4 * _ureg.meter
    nx_tomo_1.instrument.detector.set_transformation_from_x_flipped(False)
    nx_tomo_1.instrument.detector.set_transformation_from_y_flipped(True)
    nx_tomo_1.energy = 5.9 * _ureg.keV
    nx_tomo_1.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo_1.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo_1.sample.rotation_angle = numpy.linspace(0, 20, num=12) * _ureg.degree
    nx_tomo_1.sample.propagation_distance = 22.0 * _ureg.meter
    nx_tomo_1.sample.x_pixel_size = 0.023 * _ureg.millimeter
    nx_tomo_1.sample.y_pixel_size = 0.025 * _ureg.millimeter

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo_1.save(
        file_path=file_path,
        data_path=entry,
    )

    scan_1 = NXtomoScan(file_path, entry)

    nx_tomo_2 = NXtomo()
    nx_tomo_2.instrument.source.distance = 8 * _ureg.meter
    nx_tomo_2.instrument.detector.x_pixel_size = 4.023 * _ureg.meter
    nx_tomo_2.instrument.detector.y_pixel_size = 6.025 * _ureg.meter
    nx_tomo_2.instrument.detector.field_of_view = "full"
    nx_tomo_2.instrument.detector.distance = 2.89 * _ureg.meter
    lr_flip_sum = [
        trans.transformation_values
        for trans in get_lr_flip(
            nx_tomo_1.instrument.detector.transformations.transformations
        )
    ]
    lr_flip = numpy.isclose(lr_flip_sum, 180.0)
    nx_tomo_2.instrument.detector.set_transformation_from_x_flipped(not lr_flip)
    ud_flip_sum = [
        trans.transformation_values
        for trans in get_ud_flip(
            nx_tomo_1.instrument.detector.transformations.transformations
        )
    ]
    ud_flip = numpy.isclose(ud_flip_sum, 180.0)
    nx_tomo_2.instrument.detector.set_transformation_from_y_flipped(not ud_flip)
    nx_tomo_2.energy = 5.754 * _ureg.keV
    nx_tomo_2.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo_2.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo_2.sample.rotation_angle = numpy.linspace(0, 20, num=12) * _ureg.degree

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0001"
    nx_tomo_2.save(
        file_path=file_path,
        data_path=entry,
    )

    scan_2 = NXtomoScan(file_path, entry)

    # 2.0 create the widget and do the edition
    widget = NXtomoEditor()
    widget.setScan(scan=scan_1)

    for lockerButton in widget._lockerPBs:
        lockerButton.setLock(True)

    widget.setScan(scan=scan_2)
    # widget values must be the same (NXtomo field value not loaded if the lockers are active)
    assert widget._energyEntry.getValue() == 5.9
    assert widget._xDetectorPixelSizeMetricEntry.getValue() == 0.023 * _ureg.meter
    assert widget._yDetectorPixelSizeMetricEntry.getValue() == 0.025 * _ureg.meter
    assert widget._sampleDetectorDistanceMetricEntry.getValue() == 2.4 * _ureg.meter
    assert widget._propagationDistanceMetricEntry.getValue() == 22.0 * _ureg.meter
    assert widget._fieldOfViewCB.currentText() == "Full"
    assert not widget._lrFlippedCB.isChecked()
    assert widget._udFlippedCB.isChecked()
    assert widget._xSamplePixelSizeMetricEntry.getValue() == 0.023 * _ureg.millimeter
    assert widget._ySamplePixelSizeMetricEntry.getValue() == 0.025 * _ureg.millimeter
    assert widget._sampleSourceDistanceMetricEntry.getValue() == 54.8 * _ureg.meter

    # 3.0 save the nxtomo
    task = NXtomoEditorTask(
        inputs={
            "data": scan_2,
            "configuration": widget.getConfigurationForTask(),
        }
    )
    task.run()

    # 4.0 check save went well
    overwrite_nx_tomo = NXtomo().load(
        file_path=file_path,
        data_path=entry,
    )
    assert (
        overwrite_nx_tomo.instrument.detector.x_pixel_size
        == nx_tomo_1.instrument.detector.x_pixel_size
    )
    assert (
        overwrite_nx_tomo.instrument.detector.y_pixel_size
        == nx_tomo_1.instrument.detector.y_pixel_size
    )
    assert (
        overwrite_nx_tomo.instrument.detector.field_of_view
        == nx_tomo_1.instrument.detector.field_of_view
    )
    assert (
        overwrite_nx_tomo.instrument.detector.distance
        == nx_tomo_1.instrument.detector.distance
    )
    assert is_lr_flip(
        overwrite_nx_tomo.instrument.detector.transformations.transformations
    ) == is_lr_flip(nx_tomo_1.instrument.detector.transformations.transformations)
    assert is_ud_flip(
        overwrite_nx_tomo.instrument.detector.transformations.transformations
    ) == is_ud_flip(nx_tomo_1.instrument.detector.transformations.transformations)

    assert overwrite_nx_tomo.energy == nx_tomo_1.energy
    assert (
        overwrite_nx_tomo.sample.propagation_distance
        == nx_tomo_1.sample.propagation_distance
    )
    assert overwrite_nx_tomo.sample.x_pixel_size == nx_tomo_1.sample.x_pixel_size
    assert overwrite_nx_tomo.sample.y_pixel_size == nx_tomo_1.sample.y_pixel_size
    assert (
        overwrite_nx_tomo.instrument.source.distance
        == nx_tomo_1.instrument.source.distance
    )
    assert widget.getConfiguration() == {
        "instrument.beam.energy": (5.9, True),
        "instrument.detector.distance": (2.4, True),
        "instrument.detector.field_of_view": ("Full", True),
        "instrument.detector.x_pixel_size": (0.023, True),
        "instrument.detector.y_pixel_size": (0.025, True),
        "instrument.detector.lr_flipped": (False, True),
        "instrument.detector.ud_flipped": (True, True),
        "instrument.source.distance": (54.8, True),
        "sample.propagation_distance": (22.0, True),
        "sample.x_pixel_size": (2.3e-5, True),
        "sample.y_pixel_size": (2.5e-5, True),
        "sample.x_translation": (None,),
        "sample.z_translation": (None,),
    }

    for lockerButton in widget._lockerPBs:
        lockerButton.setLock(False)

    assert widget.getConfiguration() == {
        "instrument.beam.energy": (5.9, False),
        "instrument.detector.distance": (2.4, False),
        "instrument.detector.field_of_view": ("Full", False),
        "instrument.detector.x_pixel_size": (0.023, False),
        "instrument.detector.y_pixel_size": (0.025, False),
        "instrument.detector.lr_flipped": (False, False),
        "instrument.detector.ud_flipped": (True, False),
        "instrument.source.distance": (54.8, False),
        "sample.x_translation": (None,),
        "sample.z_translation": (None,),
        "sample.propagation_distance": (22.0, False),
        "sample.x_pixel_size": (2.3e-5, False),
        "sample.y_pixel_size": (2.5e-5, False),
        "sample.x_translation": (None,),
        "sample.z_translation": (None,),
    }


def test_nxtomo_editor_with_missing_paths(
    tmp_path,
    qtapp,  # noqa F811
):
    """
    test widget behavior in the case some nxtomo path don't exist
    """

    # create nx tomo with raw data
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo.sample.rotation_angle = numpy.linspace(0, 20, num=12) * _ureg.degree

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo.save(
        file_path=file_path,
        data_path=entry,
    )
    # delete some path that can be missing in some case
    with h5py.File(file_path, mode="a") as h5f:
        assert "entry0000" in h5f
        assert "entry0000/beam" not in h5f
        assert "entry0000/instrument/beam" not in h5f
        assert "entry0000/instrument/detector/distance" not in h5f
        assert "entry0000/instrument/detector/x_pixel_size" not in h5f
        assert "entry0000/instrument/detector/y_pixel_size" not in h5f
        assert "entry0000/instrument/detector/transformations" not in h5f

    scan = NXtomoScan(file_path, entry)

    # create the widget and do the edition
    widget = NXtomoEditor()

    widget.setScan(scan=scan)

    widget._sampleDetectorDistanceMetricEntry.setValue(0.05)
    widget._energyEntry.setValue(50.0)
    widget._xDetectorPixelSizeMetricEntry.setValue(0.02)
    widget._yDetectorPixelSizeMetricEntry.setValue(0.03)

    # overwrite the NXtomo
    task = NXtomoEditorTask(
        inputs={
            "data": scan,
            "configuration": widget.getConfigurationForTask(),
        }
    )
    task.run()

    # check save went well
    overwrite_nx_tomo = NXtomo().load(
        file_path=file_path,
        data_path=entry,
    )
    assert overwrite_nx_tomo.instrument.detector.x_pixel_size == 0.02 * _ureg.meter
    assert overwrite_nx_tomo.instrument.detector.y_pixel_size == 0.03 * _ureg.meter
    assert overwrite_nx_tomo.energy == 50 * _ureg.keV
    assert overwrite_nx_tomo.instrument.detector.distance == 0.05 * _ureg.meter



# coding: utf-8
from __future__ import annotations


import gc
import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.debugtools.ObjectInspectorOW import ObjectInspectorOW
from tomwer.core.utils.scanutils import MockNXtomo


class TestObjectInspectorOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        gc.collect()

    def test(self):
        widget = ObjectInspectorOW()
        widget.setObject(self.scan)
        widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        widget.close()



from __future__ import annotations

import logging

from orangewidget import gui, widget
from processview.gui.processmanager import ProcessManagerWindow

logger = logging.getLogger(__name__)


class AdvancementOW(widget.OWBaseWidget, openclass=True):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registred on the documentation
    name = "advancement"
    id = "orangecontrib.widgets.tomwer.control.AdvancementOW.AdvancementOW"
    description = "This widget can display advancement of processes and scans"
    icon = "icons/advancement.svg"
    priority = 5
    keywords = ["tomography", "process", "advancement"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)

        self._widget = ProcessManagerWindow(parent=self)

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)



from __future__ import annotations

import logging
import os
import shutil
import tempfile
import functools

from nabu.stitching.config import StitchingType
from nabu.pipeline.config import generate_nabu_configfile, parse_nabu_config_file
from nabu.stitching.config import (
    get_default_stitching_config,
    SECTIONS_COMMENTS as _SECTIONS_COMMENTS,
)

from silx.gui import qt

from tomwer.core.scan.nxtomoscan import NXtomoScan, NXtomoScanIdentifier
from tomwer.core.volume.hdf5volume import HDF5Volume, HDF5VolumeIdentifier
from tomwer.gui.qconfigfile import QConfigFileDialog
from tomwer.gui.stitching.config.positionoveraxis import PosEditorOverOneAxis
from tomwer.gui.stitching.config.output import StitchingOutput
from tomwer.gui.stitching.StitchingOptionsWidget import StitchingOptionsWidget
from tomwer.gui.configuration.action import (
    BasicConfigurationAction,
    ExpertConfigurationAction,
    MinimalisticConfigurationAction,
)
from tomwer.gui.configuration.level import ConfigurationLevel
from tomwer.gui.stitching import action as stitching_action
from tomwer.gui.stitching.preview import PreviewThread
from tomwer.gui.stitching.singleaxis import SingleAxisMetaClass, _SingleAxisMixIn
from tomwer.gui.stitching.SingleAxisStitchingWidget import SingleAxisStitchingWidget

from .utils import concatenate_dict

_logger = logging.getLogger(__name__)


class _SingleAxisStitchingWindow(
    qt.QMainWindow, _SingleAxisMixIn, metaclass=SingleAxisMetaClass
):
    """
    Main widget containing all the options to define the stitching to be done

    :param with_configuration_action: if True append the load and save stitching configuration tool button.
                                           In some cases those can also be part of Menu so we want to avoid having those twice
    """

    sigChanged = qt.Signal()
    """Signal emit each time the configuration is modified"""

    def __init__(self, parent=None, with_configuration_action=True) -> None:
        super().__init__(parent)
        assert self._axis in (0, 1, 2)
        """axis along which we want to apply stitching"""
        self._previewFolder = None
        # folder to store files (volume or NXtomo) for previews
        self._previewThread = None
        # thread to compute the stitching for preview
        self._callbackToGetSlurmConfig = None
        self._callbackToSetSlurmConfig = None
        # convenient work around to avoid having to redefine the n=interface for slurm and the API
        # to load and save settings
        # if it is defined upper

        toolbar = qt.QToolBar(self)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)
        style = qt.QApplication.instance().style()

        # clean option
        self.__cleanAction = qt.QAction(self)
        self.__cleanAction.setToolTip("clear")
        clear_icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        self.__cleanAction.setIcon(clear_icon)
        toolbar.addAction(self.__cleanAction)
        self.__cleanAction.triggered.connect(self.clean)

        # separator
        toolbar.addSeparator()

        if with_configuration_action:
            # load action
            self.__loadAction = stitching_action.LoadConfigurationAction(self)
            toolbar.addAction(self.__loadAction)
            self.__loadAction.triggered.connect(
                functools.partial(self._loadSettings, file_path=None)
            )

            # save action
            self.__saveAction = stitching_action.SaveConfigurationAction(self)
            toolbar.addAction(self.__saveAction)
            self.__saveAction.triggered.connect(
                functools.partial(self._saveSettings, file_path=None)
            )

            # separator
            toolbar.addSeparator()

        # update preview action
        self.__updatePreviewAction = stitching_action.PreviewAction(self)
        toolbar.addAction(self.__updatePreviewAction)
        self.__updatePreviewAction.triggered.connect(self._trigger_update_preview)

        # separator
        toolbar.addSeparator()

        # configuration level / mode
        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        self._minimalisticAction = MinimalisticConfigurationAction(toolbar)
        menu.addAction(self._minimalisticAction)
        self.__configurationModesGroup.addAction(self._minimalisticAction)
        self._basicConfigAction = BasicConfigurationAction(toolbar)
        menu.addAction(self._basicConfigAction)
        self.__configurationModesGroup.addAction(self._basicConfigAction)
        self._expertConfiguration = ExpertConfigurationAction(toolbar)
        menu.addAction(self._expertConfiguration)
        self.__configurationModesGroup.addAction(self._expertConfiguration)

        # separator
        toolbar.addSeparator()

        # create central widget
        self._widget = SingleAxisStitchingWidget(parent=self, axis=self.first_axis)
        self.setCentralWidget(self._widget)

        # create Dock widgets
        ##  output
        self._outputWidget = StitchingOutput(parent=self)
        self._outputWidget.setObjectName("outputSettingsWidget")
        self._outputDockWidget = qt.QDockWidget(parent=self)
        self._outputDockWidget.setWindowTitle("output")
        self._outputDockWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._outputDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._outputDockWidget.setWidget(self._outputWidget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._outputDockWidget)
        self._outputDockWidget.setToolTip(
            "options to where and how to save the stitching"
        )
        ##  stitching strategies
        self._stitchingOptsWidget = StitchingOptionsWidget(
            parent=self, first_axis=self.first_axis, second_axis=self.second_axis
        )
        self._stitchingOptsScrollArea = qt.QScrollArea(self)
        self._stitchingOptsScrollArea.setWidget(self._stitchingOptsWidget)
        self._stitchingOptsScrollArea.setWidgetResizable(True)
        self._stitchingOptsScrollArea.setHorizontalScrollBarPolicy(
            qt.Qt.ScrollBarAlwaysOff
        )
        self._stitchingOptsDockWidget = qt.QDockWidget(parent=self)
        self._stitchingOptsDockWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._stitchingOptsDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._stitchingOptsDockWidget.setWidget(self._stitchingOptsScrollArea)
        self._stitchingOptsDockWidget.setWindowTitle("processing options")
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._stitchingOptsDockWidget)

        ## all scan z positions
        self._editTomoObjFirstAxisPositionsWidget = PosEditorOverOneAxis(
            parent=self,
            axis_edited=self.first_axis,
            axis_order=self.first_axis,
        )
        self._editTomoObjFirstAxisPositionsDockWidget = qt.QDockWidget(parent=self)
        self._editTomoObjFirstAxisPositionsDockWidget.layout().setContentsMargins(
            0, 0, 0, 0
        )
        self._editTomoObjFirstAxisPositionsDockWidget.setFeatures(
            qt.QDockWidget.DockWidgetMovable
        )
        self._editTomoObjFirstAxisPositionsDockWidget.setWidget(
            self._editTomoObjFirstAxisPositionsWidget
        )
        self._editTomoObjFirstAxisPositionsDockWidget.setWindowTitle(
            f"edit positions over axis {self.first_axis} (px) - aka {self.axis_alias(self.first_axis)}"
        )
        self._editTomoObjFirstAxisPositionsDockWidget.setToolTip(
            f"This allows to edit tomo objects positions along the axis {self.first_axis} (aka {self.axis_alias(self.first_axis)})"
        )
        self.addDockWidget(
            qt.Qt.RightDockWidgetArea, self._editTomoObjFirstAxisPositionsDockWidget
        )
        ### add a check box to update position from preview if asked by the user
        self._updateFirstAxisPosFromPreviewCalc = qt.QCheckBox(
            f"update position {self.first_axis} from preview calc",
            self,
        )
        self._updateFirstAxisPosFromPreviewCalc.setToolTip(
            f"When the user trigger a preview, if some shift search refined over axis {self.first_axis} is done then will update the axis 0 positions",
        )
        self._updateFirstAxisPosFromPreviewCalc.setChecked(True)
        self._editTomoObjFirstAxisPositionsWidget.layout().insertWidget(
            0, self._updateFirstAxisPosFromPreviewCalc
        )

        ## all scan axis 1 positions
        self._editTomoObjSecondAxisPositionsWidget = PosEditorOverOneAxis(
            parent=self,
            axis_edited=self.second_axis,
            axis_order=0,
        )
        self._editTomoObjSecondAxisPositionsDockWidget = qt.QDockWidget(parent=self)
        self._editTomoObjSecondAxisPositionsDockWidget.layout().setContentsMargins(
            0, 0, 0, 0
        )
        self._editTomoObjSecondAxisPositionsDockWidget.setFeatures(
            qt.QDockWidget.DockWidgetMovable
        )
        self._editTomoObjSecondAxisPositionsDockWidget.setWidget(
            self._editTomoObjSecondAxisPositionsWidget
        )
        self._editTomoObjSecondAxisPositionsDockWidget.setWindowTitle(
            f"edit positions over axis {self.second_axis} (px)"
        )
        self._editTomoObjSecondAxisPositionsDockWidget.setToolTip(
            f"This allows to edit tomo objects positions along the axis {self.second_axis}"
        )
        self.addDockWidget(
            qt.Qt.RightDockWidgetArea, self._editTomoObjSecondAxisPositionsDockWidget
        )
        ### add a check box to update position from preview if asked by the user
        self._updateSecondAxisPosFromPreviewCalc = qt.QCheckBox(
            f"update position {self.second_axis} from preview calc", self
        )
        self._updateSecondAxisPosFromPreviewCalc.setToolTip(
            f"When the user trigger a preview, if some position refinement can be done will update axis {self.second_axis}"
        )
        self._updateSecondAxisPosFromPreviewCalc.setChecked(True)
        self._editTomoObjSecondAxisPositionsWidget.layout().insertWidget(
            0, self._updateSecondAxisPosFromPreviewCalc
        )

        self._widget.setAddTomoObjCallbacks(
            (
                self._editTomoObjFirstAxisPositionsWidget.addTomoObj,
                self._editTomoObjSecondAxisPositionsWidget.addTomoObj,
                self.getRawDisplayPlot().addTomoObj,
            )
        )
        self._widget.setRemoveTomoObjCallbacks(
            (
                self._editTomoObjFirstAxisPositionsWidget.removeTomoObj,
                self._editTomoObjSecondAxisPositionsWidget.removeTomoObj,
                self.getRawDisplayPlot().removeTomoObj,
            )
        )

        # update layout: for now lets tabify some widget
        self.tabifyDockWidget(self._outputDockWidget, self._stitchingOptsDockWidget)
        self.tabifyDockWidget(
            self._outputDockWidget, self._editTomoObjSecondAxisPositionsDockWidget
        )
        self.tabifyDockWidget(
            self._outputDockWidget, self._editTomoObjFirstAxisPositionsDockWidget
        )

        # handle raw display plot. By display avoid displaying raw data as this can be resource consuming
        self._widget._mainWidget._rawDisplayCB.setChecked(False)
        self._widget._mainWidget._rawDisplayPlot.setActive(False)

        # connect signal / slot
        self._widget._mainWidget._rawDisplayCB.toggled.connect(
            self._handleRawDisplayConnection
        )
        self._outputWidget.sigChanged.connect(self._changed)
        self._stitchingOptsWidget.sigChanged.connect(self._changed)
        self._widget.sigStitchingTypeChanged.connect(
            self._outputWidget._updateOutputForStitchingType
        )
        self._widget.sigStitchingTypeChanged.connect(
            self._stitchingOptsWidget._stitchingTypeChanged
        )

        ## handle raw plot preview
        self._stitchingOptsWidget.sigFlipLRChanged.connect(
            self.getRawDisplayPlot().setFlipLRFrames
        )
        self._stitchingOptsWidget.sigFlipUDChanged.connect(
            self.getRawDisplayPlot().setFlipUDFrames
        )
        self._stitchingOptsWidget.sigSliceForPreviewChanged.connect(
            self.getRawDisplayPlot().setSliceForPreview
        )

        ## handle tomo obj loading from settings
        self._widget.sigTomoObjsLoaded.connect(
            self._editTomoObjFirstAxisPositionsWidget.setTomoObjs
        )
        self._widget.sigTomoObjsLoaded.connect(
            self._editTomoObjSecondAxisPositionsWidget.setTomoObjs
        )
        self._widget.sigTomoObjsLoaded.connect(self.getRawDisplayPlot().setTomoObjs)

        # set up
        self._basicConfigAction.setChecked(True)
        self._userModeChanged(self._basicConfigAction)

    def setCallbackToGetSlurmConfig(self, callback):
        self._callbackToGetSlurmConfig = callback

    def setCallbackToSetSlurmConfig(self, callback):
        self._callbackToSetSlurmConfig = callback

    def close(self):
        # remove folder used for preview
        if self._previewFolder is not None:
            shutil.rmtree(self._previewFolder, ignore_errors=True)
        self._widget.close()
        # requested for the waiting plot update
        super().close()

    def getRawDisplayPlot(self):
        return self._widget._mainWidget._rawDisplayPlot

    def _handleRawDisplayConnection(self, toggled: bool):
        raw_display_plot = self.getRawDisplayPlot()
        raw_display_plot.setActive(toggled)

    def getPreviewAction(self):
        return self.__updatePreviewAction

    def getPreviewFolder(self):
        if self._previewFolder is None:
            self._previewFolder = tempfile.mkdtemp(prefix="tomwer_stitcher_preview")
        return self._previewFolder

    def getVolumeIdentifierPreview(self) -> HDF5VolumeIdentifier:
        folder = self.getPreviewFolder()
        # for now use hdf5 by default
        return HDF5VolumeIdentifier(
            object=HDF5Volume,
            hdf5_file=os.path.join(folder, "vol_stitching_preview.hdf5"),
            entry="my_volume",
        )

    def getNXtomoIdentifierForPreview(self):
        folder = self.getPreviewFolder()
        return NXtomoScanIdentifier(
            object=NXtomoScan,
            hdf5_file=os.path.join(folder, "nxtomo_stiching_preview.hdf5"),
            entry="entry0000",
        )

    def _changed(self, *args, **kwargs):
        self.sigChanged.emit()

    def _saveSettings(self, file_path=None, **kwargs):
        """
        dump current configuration into a txt file
        """
        # get a file if necessary
        if file_path is None:
            dialog = QConfigFileDialog(self)
            dialog.setAcceptMode(qt.QFileDialog.AcceptSave)
            if not dialog.exec():
                return

            selected_file = dialog.selectedFiles()
            if len(selected_file) == 0:
                return
            file_path = selected_file[0]

        configuration = self.getConfiguration()
        if self._callbackToGetSlurmConfig is not None:
            slurm_config = {"slurm": self._callbackToGetSlurmConfig()}
            configuration = concatenate_dict(configuration, slurm_config)

        _, ext = os.path.splitext(file_path)
        if ext == "":
            file_path = f"{file_path}.cfg"
        # dump configuration
        generate_nabu_configfile(
            fname=file_path,
            default_config=get_default_stitching_config(self.getStitchingType()),
            comments=True,
            sections_comments=_SECTIONS_COMMENTS,
            options_level="advanced",
            prefilled_values=configuration,
        )

    def _loadSettings(self, file_path=None, **kwargs):
        """
        load configuration from a txt file
        """
        # get a file if necessary
        if file_path is None:
            dialog = QConfigFileDialog(self)
            dialog.setAcceptMode(qt.QFileDialog.AcceptOpen)
            dialog.setFileMode(qt.QFileDialog.ExistingFiles)

            if not dialog.exec():
                return

            selected_file = dialog.selectedFiles()
            if len(selected_file) == 0:
                return
            file_path = selected_file[0]

        # Do configuration load
        conf_dict = parse_nabu_config_file(file_path, allow_no_value=True)
        self.setConfiguration(config=conf_dict)
        if self._callbackToSetSlurmConfig is not None:
            self._callbackToSetSlurmConfig(conf_dict.get("slurm", {}))

    def _trigger_update_preview(self):
        if self._previewThread is not None:
            _logger.warning(
                "some preview is already running. Please wait before relaunching it"
            )
            return
        config = self.getConfiguration()
        # update the output file to set if from raw...
        stitching_type = config.get("stitching", {}).get("type", None)
        if stitching_type in ("z-preproc", "y-preproc"):
            output_identifier = self.getNXtomoIdentifierForPreview()
            config["preproc"]["location"] = output_identifier.file_path
            config["preproc"]["data_path"] = output_identifier.data_path
            assert "postproc" not in config
        elif stitching_type == "z-postproc":
            config["postproc"][
                "output_volume"
            ] = self.getVolumeIdentifierPreview().to_str()
            assert "preproc" not in config
        else:
            raise NotImplementedError

        # update the slice to avoid doing the stitching on all the frames
        config["inputs"]["slices"] = self.getSlicesForPreview()

        # update to force overwrite
        config["output"]["overwrite_results"] = True

        # avoid any recalculation of the axis position. So the preview is accurate
        config["output"]["axis_0_params"] = ""
        config["output"]["axis_1_params"] = ""
        config["output"]["axis_2_params"] = ""

        # clean current preview to notify some calculation is going on
        preview_plot = self._widget._mainWidget._previewPlot
        preview_plot._waitingOverlay.show()

        # start stitching on a thread
        self._previewThread = PreviewThread(stitching_config=config)
        self._previewThread.finished.connect(self._previewCalculationFinished)
        self._previewThread.start()

    def getSlicesForPreview(self):
        return self._stitchingOptsWidget.getSlicesForPreview()

    def _previewCalculationFinished(self):
        sender = self.sender()
        assert isinstance(sender, PreviewThread)
        composition = sender.frame_composition
        tomo_objs_new_axis_positions = sender.final_tomo_objs_positions
        assert isinstance(
            tomo_objs_new_axis_positions, dict
        ), f"final_tomo_objs_positions is expected to be a dict with obj identifier as key and the tuple of position as value. Got {type(tomo_objs_new_axis_positions)}"
        # expect it to be a dict with tomo obj identifier as key and a tuple of (axis_2_pos, axis_1_pos, axis_0_pos) as value
        output_obj_identifier = sender.output_identifier

        preview_plot = self._widget._mainWidget._previewPlot
        preview_plot._waitingOverlay.hide()

        self._previewThread.finished.disconnect(self._previewCalculationFinished)
        self._previewThread = None

        if output_obj_identifier is None:
            _logger.error("preview of stitching failed")
        else:
            preview_plot.setStitchedTomoObj(
                tomo_obj_id=output_obj_identifier,
                composition=composition,
            )

        # update object values if requested
        update_requested = {
            self.first_axis: self._updateFirstAxisPosFromPreviewCalc.isChecked(),
            self.second_axis: self._updateSecondAxisPosFromPreviewCalc.isChecked(),
        }

        if update_requested[self.first_axis] or update_requested[self.second_axis]:
            existing_tomo_obj = {
                tomo_obj.get_identifier().to_str(): tomo_obj
                for tomo_obj in self._widget._mainWidget.getTomoObjs()
            }

            for tomo_obj_id, value in tomo_objs_new_axis_positions.items():
                assert (
                    isinstance(value, tuple) and len(value) == 3
                ), "value is expected to be (new_pos_axis_0, new_pos_axis_1, new_pos_axis_2)"
                new_first_axis_pos = value[self.first_axis]
                new_second_axis_pos = value[self.second_axis]
                tomo_obj = existing_tomo_obj.get(tomo_obj_id, None)
                if tomo_obj is None:
                    continue

                if update_requested[self.first_axis]:
                    tomo_obj.stitching_metadata.setPxPos(
                        int(new_first_axis_pos), self.first_axis
                    )
                if update_requested[self.second_axis]:
                    tomo_obj.stitching_metadata.setPxPos(
                        int(new_second_axis_pos), self.second_axis
                    )
            if update_requested[self.first_axis]:
                self._editTomoObjFirstAxisPositionsWidget._orderedMightHaveChanged(
                    force_sb_update=True
                )
            if update_requested[self.second_axis]:
                self._editTomoObjSecondAxisPositionsWidget._orderedMightHaveChanged(
                    force_sb_update=True
                )

    def clean(self):
        self._widget.clean()
        self._editTomoObjFirstAxisPositionsWidget.clean()
        self._editTomoObjSecondAxisPositionsWidget.clean()

    def setSeries(self, series):
        self.clean()
        self._widget._mainWidget.setSeries(series)
        self._editTomoObjFirstAxisPositionsWidget.clean()
        self._editTomoObjSecondAxisPositionsWidget.clean()
        for tomo_obj in series:
            self._editTomoObjFirstAxisPositionsWidget.addTomoObj(tomo_obj)
            self._editTomoObjSecondAxisPositionsWidget.addTomoObj(tomo_obj)
        self.getRawDisplayPlot().setTomoObjs(tomo_objs=series[:])

    def addTomoObj(self, tomo_obj):
        self._widget.addTomoObj(tomo_obj)
        self._editTomoObjFirstAxisPositionsWidget.addTomoObj(tomo_obj)
        self._editTomoObjSecondAxisPositionsWidget.addTomoObj(tomo_obj)
        self.getRawDisplayPlot().addTomoObj(tomo_obj=tomo_obj)

    def getTomoObjs(self) -> tuple:
        return self._widget.getTomoObjs()

    def removeTomoObj(self, tomo_obj):
        self._widget.removeTomoObj(tomo_obj)
        self.getRawDisplayPlot().removeTomoObj(tomo_obj=tomo_obj)

    def getConfiguration(self) -> dict:
        # make sure the sync is fine between the two
        configs = (
            self._widget.getConfiguration(),
            self._outputWidget.getConfiguration(),
            self._stitchingOptsWidget.getConfiguration(),
        )
        result = {}
        for config in configs:
            result = concatenate_dict(result, config)
        return result

    def setConfiguration(self, config: dict):
        self._widget.setConfiguration(config)
        self._outputWidget.setConfiguration(config)
        self._stitchingOptsWidget.setConfiguration(config)

    # expose API
    def getStitchingType(self) -> StitchingType:
        return self._widget.getStitchingType()

    def setStitchingType(self, stitching_type: StitchingType):
        self._widget.setStitchingType(stitching_type)

    def _userModeChanged(self, action):
        self.__configurationModesAction.setIcon(action.icon())
        self.__configurationModesAction.setToolTip(action.tooltip())
        if action is self._basicConfigAction:
            level = ConfigurationLevel.OPTIONAL
        elif action is self._expertConfiguration:
            level = ConfigurationLevel.ADVANCED
        else:
            level = ConfigurationLevel.REQUIRED
        self._stitchingOptsWidget.setConfigurationLevel(level)
        self._editTomoObjSecondAxisPositionsDockWidget.setVisible(
            level >= ConfigurationLevel.ADVANCED
        )

    def setPreProcessingOutput(self, *args, **kwargs):
        self._outputWidget.setPreProcessingOutput(*args, **kwargs)

    def setPostProcessingOutput(self, *args, **kwargs):
        self._outputWidget.setPostProcessingOutput(*args, **kwargs)

    def get_available_pre_processing_stitching_mode(self):
        return self._widget.get_available_pre_processing_stitching_mode()

    def get_available_post_processing_stitching_mode(self):
        return self._widget.get_available_post_processing_stitching_mode()


class YStitchingWindow(_SingleAxisStitchingWindow, axis=1):
    pass


class ZStitchingWindow(_SingleAxisStitchingWindow, axis=0):
    pass



# coding: utf-8
from __future__ import annotations

import logging
import weakref

import numpy
from silx.gui import icons as silx_icons
from silx.gui import qt
from silx.gui.plot.CompareImages import (
    CompareImages as _CompareImages,
    VisualizationMode,
)

from silx.gui.plot.tools.compare.toolbar import VisualizationModeToolButton
from silx.gui.plot import tools


_logger = logging.getLogger(__name__)


class CompareImages(_CompareImages):
    sigCropImagesChanged = qt.Signal()
    """Emit when cropping of the compared images has changed"""

    def __init__(self, parent=None, backend=None):
        super().__init__(parent, backend)

        # create crop images action
        icon = silx_icons.getQIcon("crop")
        action = qt.QAction(icon, "crop compared images", self)
        action.setCheckable(True)
        action.setChecked(True)
        action.triggered.connect(self.__cropComparedImagesChanged)
        self.__cropComparedImages = action
        self._compareToolBar.addAction(action)

        # define visible compare mode
        self._compareToolBar._visualizationToolButton.setVisibleModes(
            [
                VisualizationMode.ONLY_A,
                VisualizationMode.ONLY_B,
                VisualizationMode.COMPOSITE_RED_BLUE_GRAY_NEG,
                VisualizationMode.COMPOSITE_A_MINUS_B,
            ]
        )

    def __cropComparedImagesChanged(self):
        cropCompositeImage = self.__cropComparedImages.isChecked()
        self.setCropComparedImages(cropCompositeImage)

    def cropComparedImages(self) -> bool:
        return self.__cropComparedImages.isChecked()

    def setCropComparedImages(self, crop):
        self.__cropComparedImages.setChecked(crop)
        self.sigCropImagesChanged.emit()

    def _createToolBars(self, plot):
        """Create tool bars displayed by the widget"""
        toolBar = tools.InteractiveModeToolBar(parent=self, plot=plot)
        self._interactiveModeToolBar = toolBar
        toolBar = tools.ImageToolBar(parent=self, plot=plot)
        self._imageToolBar = toolBar
        toolBar = CompareImagesToolBar(self)
        toolBar.setCompareWidget(self)
        self._compareToolBar = toolBar


class CompareImagesStatusBar(qt.QStatusBar):
    """StatusBar containing specific information contained in a
    :class:`CompareImages` widget

    Use :meth:`setCompareWidget` to connect this toolbar to a specific
    :class:`CompareImages` widget.

    :param parent: Parent of this widget.
    """

    def __init__(self, parent=None):
        qt.QStatusBar.__init__(self, parent)
        self.setSizeGripEnabled(False)
        self.layout().setSpacing(0)
        self.__compareWidget = None
        self._label1 = qt.QLabel(self)
        self._label1.setFrameShape(qt.QFrame.WinPanel)
        self._label1.setFrameShadow(qt.QFrame.Sunken)
        self._label2 = qt.QLabel(self)
        self._label2.setFrameShape(qt.QFrame.WinPanel)
        self._label2.setFrameShadow(qt.QFrame.Sunken)
        self._transform = qt.QLabel(self)
        self._transform.setFrameShape(qt.QFrame.WinPanel)
        self._transform.setFrameShadow(qt.QFrame.Sunken)
        self.addWidget(self._label1)
        self.addWidget(self._label2)
        self.addWidget(self._transform)
        self._pos = None
        self._updateStatusBar()

    def setCompareWidget(self, widget):
        """
        Connect this tool bar to a specific :class:`CompareImages` widget.

        :param widget: The widget to connect with.
        """
        compareWidget = self.getCompareWidget()
        if compareWidget is not None:
            compareWidget.getPlot().sigPlotSignal.disconnect(self.__plotSignalReceived)
            compareWidget.sigConfigurationChanged.disconnect(self.__dataChanged)
        compareWidget = widget
        if compareWidget is None:
            self.__compareWidget = None
        else:
            self.__compareWidget = weakref.ref(compareWidget)
        if compareWidget is not None:
            compareWidget.getPlot().sigPlotSignal.connect(self.__plotSignalReceived)
            compareWidget.sigConfigurationChanged.connect(self.__dataChanged)

    def getCompareWidget(self) -> CompareImages:
        """Returns the connected widget."""
        if self.__compareWidget is None:
            return None
        else:
            return self.__compareWidget()

    def __plotSignalReceived(self, event):
        """Called when old style signals at emmited from the plot."""
        if event["event"] == "mouseMoved":
            x, y = event["x"], event["y"]
            self.__mouseMoved(x, y)

    def __mouseMoved(self, x, y):
        """Called when mouse move over the plot."""
        self._pos = x, y
        self._updateStatusBar()

    def __dataChanged(self):
        """Called when internal data from the connected widget changes."""
        self._updateStatusBar()

    def _formatData(self, data: int | float | None) -> str:
        """Format pixel of an image.

        It supports intensity, RGB, and RGBA.

        :param data: Value of a pixel
        """
        if data is None:
            return "-"
        if isinstance(data, (int, numpy.integer)):
            return "%d" % data
        if isinstance(data, (float, numpy.floating)):
            return "%f" % data
        if isinstance(data, numpy.ndarray):
            # RGBA value
            if data.shape == (3,):
                return "R:%d G:%d B:%d" % (data[0], data[1], data[2])
            elif data.shape == (4,):
                return "R:%d G:%d B:%d A:%d" % (data[0], data[1], data[2], data[3])
        _logger.debug("Unsupported data format %s. Cast it to string.", type(data))
        return str(data)

    def _updateStatusBar(self):
        """Update the content of the status bar"""
        widget = self.getCompareWidget()
        if widget is None:
            self._label1.setText("Frame 1: -")
            self._label2.setText("Frame 2: -")
            self._transform.setVisible(False)
        else:
            transform = widget.getTransformation()
            self._transform.setVisible(transform is not None)
            if transform is not None:
                has_notable_translation = not numpy.isclose(
                    transform.tx, 0.0, atol=0.01
                ) or not numpy.isclose(transform.ty, 0.0, atol=0.01)
                has_notable_scale = not numpy.isclose(
                    transform.sx, 1.0, atol=0.01
                ) or not numpy.isclose(transform.sy, 1.0, atol=0.01)
                has_notable_rotation = not numpy.isclose(transform.rot, 0.0, atol=0.01)

                strings = []
                if has_notable_translation:
                    strings.append("Translation")
                if has_notable_scale:
                    strings.append("Scale")
                if has_notable_rotation:
                    strings.append("Rotation")
                if strings == []:
                    has_translation = not numpy.isclose(
                        transform.tx, 0.0
                    ) or not numpy.isclose(transform.ty, 0.0)
                    has_scale = not numpy.isclose(
                        transform.sx, 1.0
                    ) or not numpy.isclose(transform.sy, 1.0)
                    has_rotation = not numpy.isclose(transform.rot, 0.0)
                    if has_translation or has_scale or has_rotation:
                        text = "No big changes"
                    else:
                        text = "No changes"
                else:
                    text = "+".join(strings)
                self._transform.setText("Align: " + text)

                strings = []
                if not numpy.isclose(transform.ty, 0.0):
                    strings.append("Translation x: %0.3fpx" % transform.tx)
                if not numpy.isclose(transform.ty, 0.0):
                    strings.append("Translation y: %0.3fpx" % transform.ty)
                if not numpy.isclose(transform.sx, 1.0):
                    strings.append("Scale x: %0.3f" % transform.sx)
                if not numpy.isclose(transform.sy, 1.0):
                    strings.append("Scale y: %0.3f" % transform.sy)
                if not numpy.isclose(transform.rot, 0.0):
                    _rot = transform.rot * 180 / numpy.pi
                    strings.append(f"Rotation: {_rot:0.3f}deg")
                if strings == []:
                    text = "No transformation"
                else:
                    text = "\n".join(strings)
                self._transform.setToolTip(text)

            if self._pos is None:
                self._label1.setText("Frame 1: -")
                self._label2.setText("Frame 2: -")
            else:
                data1, data2 = widget.getRawPixelData(self._pos[0], self._pos[1])
                if isinstance(data1, str):
                    self._label1.setToolTip(data1)
                    text1 = "-"
                else:
                    self._label1.setToolTip("")
                    text1 = self._formatData(data1)
                if isinstance(data2, str):
                    self._label2.setToolTip(data2)
                    text2 = "-"
                else:
                    self._label2.setToolTip("")
                    text2 = self._formatData(data2)
                self._label1.setText("Frame 1: %s" % text1)
                self._label2.setText("Frame 2: %s" % text2)


class CompareImagesToolBar(qt.QToolBar):
    def __init__(self, parent=None):
        qt.QToolBar.__init__(self, parent)

        self.__compareWidget = None

        self._visualizationToolButton = VisualizationModeToolButton(self)
        self._visualizationToolButton.setPopupMode(qt.QToolButton.InstantPopup)
        self._visualizationToolButton.sigSelected.connect(self.__visualizationChanged)
        self.addWidget(self._visualizationToolButton)

    def __visualizationChanged(self, mode: VisualizationMode):
        widget = self.getCompareWidget()
        if widget is not None:
            widget.setVisualizationMode(mode)

    def setCompareWidget(self, widget: None | CompareImages):
        """
        Connect this tool bar to a specific :class:`CompareImages` widget.

        :param widget: The widget to connect the toolbar with.
        """
        compareWidget = self.getCompareWidget()
        if compareWidget is not None:
            compareWidget.sigConfigurationChanged.disconnect(
                self.__updateSelectedActions
            )
        compareWidget = widget
        self.setEnabled(compareWidget is not None)
        if compareWidget is None:
            self.__compareWidget = None
        else:
            self.__compareWidget = weakref.ref(compareWidget)
        if compareWidget is not None:
            widget.sigConfigurationChanged.connect(self.__updateSelectedActions)
        self.__updateSelectedActions()

    def getCompareWidget(self) -> CompareImages:
        """Returns the connected widget."""
        if self.__compareWidget is None:
            return None
        else:
            return self.__compareWidget()

    def __updateSelectedActions(self):
        """
        Update the state of this tool bar according to the state of the
        connected :class:`CompareImages` widget.
        """
        widget = self.getCompareWidget()
        if widget is None:
            return
        self._visualizationToolButton.setSelected(widget.getVisualizationMode())



from __future__ import annotations

import logging
import numpy
import pint
from functools import partial
from silx.gui.plot.items.axis import Axis
from silx.gui import qt

from tomoscan.volumebase import SliceTuple
from tomoscan.identifier import VolumeIdentifier

from tomoscan.esrf.volume.singleframebase import VolumeSingleFrameBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume import HDF5Volume
from tomwer.gui.visualization.fullscreenplot import FullScreenPlot2D
from tomwer.gui.visualization.TomoPlot2D import TomoPlot2D as _TomoPlot2D

from .GeometryOrMetadataWidget import GeometryOrMetadataWidget
from ._TomoPlot2DSlices import TomoPlot2DSlices


_logger = logging.getLogger(__name__)
_ureg = pint.get_application_registry()


class VolumeViewerWidget(qt.QWidget):
    """
    A window displaying slices side by side along the three direction with a 'geometric reference'.
    """

    _HANDLE_WIDTH = 10  # px

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setLayout(qt.QVBoxLayout())
        self._verticalSplitter = qt.QSplitter(qt.Qt.Vertical)
        self._verticalSplitter.setHandleWidth(self._HANDLE_WIDTH)
        self.layout().addWidget(self._verticalSplitter)

        # horizontal top splitter
        self._horizontalTopSplitter = qt.QSplitter(qt.Qt.Horizontal)
        self._horizontalTopSplitter.setHandleWidth(self._HANDLE_WIDTH)
        self._XYPlot2D = TomoPlot2DSlices(
            parent=self,
            title="XY",
            color="red",
            image="3D_coordinate_system_XY",
            axis=0,
        )
        self._horizontalTopSplitter.addWidget(self._XYPlot2D)

        self._YZPlot2D = TomoPlot2DSlices(
            parent=self,
            title="YZ",
            color="blue",
            image="3D_coordinate_system_YZ",
            axis=2,
        )
        self._horizontalTopSplitter.addWidget(self._YZPlot2D)
        self._verticalSplitter.addWidget(self._horizontalTopSplitter)

        # vertical bottom splitter
        self._horizontalBottomSplitter = qt.QSplitter(qt.Qt.Horizontal)
        self._horizontalBottomSplitter.setHandleWidth(self._HANDLE_WIDTH)
        self._geometryOrMetadataWidget = GeometryOrMetadataWidget(parent=self)
        self._horizontalBottomSplitter.addWidget(self._geometryOrMetadataWidget)
        self._XZPlot2D = TomoPlot2DSlices(
            parent=self,
            title="XZ",
            color="green",
            image="3D_coordinate_system_XZ",
            axis=1,
        )
        self._horizontalBottomSplitter.addWidget(self._XZPlot2D)
        self._verticalSplitter.addWidget(self._horizontalBottomSplitter)

        # define size policy
        self._YZPlot2D.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self._XZPlot2D.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self._XYPlot2D.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self._geometryOrMetadataWidget.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )

        # connect signal / slot
        for plot in (
            self._YZPlot2D,
            self._XZPlot2D,
            self._XYPlot2D,
        ):
            plot.displayFullScreen.connect(
                partial(
                    self._popUpImageFullScreen,
                    plot,
                )
            )

    def initVolumePreview(self, volume: TomwerVolumeBase, message: str | None):
        """
        Set display for a new volume before the full volume or slices are loaded.

        For HDF5 and VolumeSingleFrame it will allow users to browse the dataset before the full volume or slices subset are loaded.
        Else we notify the user that some loading is on-going.
        """
        self._geometryOrMetadataWidget.setVolumeIdentifier(volume.get_identifier())
        if isinstance(volume, (HDF5Volume, VolumeSingleFrameBase)):
            # Z is the fast read axis for most file format. So we allow users to access it directly.
            _logger.info(f"Set {volume.get_identifier()} for preview.")
            self._XYPlot2D.setVolume(volume)

            vol_shape = volume.get_volume_shape()
            if vol_shape:
                self.setVolumeShape(shape=vol_shape)
                self._XYPlot2D.setCurrentIndex(vol_shape[0] // 2)
        elif isinstance(volume, TomwerVolumeBase):
            _logger.debug(f"No preview available for {volume.get_identifier()}.")
            self._XYPlot2D.setLoadingMessage(message)
            self._XYPlot2D.setLoading(True)
        else:
            raise TypeError(f"type not handled: {type(volume)}")
        for plot in (self._XZPlot2D, self._YZPlot2D):
            plot.setLoadingMessage(message)
            plot.clear()
            plot.setLoading(True)

    def setVolume(self, volume: TomwerVolumeBase):
        """
        Defines the volume. Expects to be loaded in memory to avoid freezing the GUI when browsing slices.
        """
        self._XYPlot2D.setVolume(volume)
        self._XZPlot2D.setVolume(volume)
        self._YZPlot2D.setVolume(volume)

    def stopPreview(self):
        for plot in self._window.getPlots():
            plot.setLoading(False)

    def getPlots(self) -> dict[str, _TomoPlot2D]:
        return {
            "XY": self._XYPlot2D.getPlot(),
            "YZ": self._YZPlot2D.getPlot(),
            "XZ": self._XZPlot2D.getPlot(),
        }

    def extendPlot(self, plot_name: str | None):
        """
        Extend the requested plot to take most of the space in the widget.
        If None 'reset' the view to equally space the plots and metadata

        :param plot_name: None for reset else should be in ('XY', 'XZ', 'YZ')
        """
        # warning: the bottom slides are using 3 values because we expect to have the
        # ColorbarWidget defined by the VolumeReconstructionSummaryWindow
        # reset splitters
        if plot_name is None:
            self._horizontalTopSplitter.setSizes([1, 1])
            self._horizontalBottomSplitter.setSizes([2, 4, 6])
            self._verticalSplitter.setSizes([1, 1])
        elif plot_name == "XY":
            self._verticalSplitter.setSizes([1, 0])
            self._horizontalTopSplitter.setSizes([1, 0])
        elif plot_name == "YZ":
            self._verticalSplitter.setSizes([1, 0])
            self._horizontalTopSplitter.setSizes([0, 1])
        elif plot_name == "XZ":
            self._verticalSplitter.setSizes([0, 1])
            self._horizontalBottomSplitter.setSizes([0, 0, 1])
        else:
            raise ValueError(f"{plot_name!r} not handle.")

    def clear(self) -> None:
        self._YZPlot2D.clear()
        self._XZPlot2D.clear()
        self._XYPlot2D.clear()
        self._geometryOrMetadataWidget.clear()

    def setLoading(self, loading: bool):
        self._YZPlot2D.setLoading(loading)
        self._XZPlot2D.setLoading(loading)
        self._XYPlot2D.setLoading(loading)

    def setSlices(self, slices: dict[SliceTuple, numpy.ndarray]):
        for axis, plot in {
            0: self._XYPlot2D,
            1: self._YZPlot2D,
            2: self._XZPlot2D,
        }.items():
            plot.setSlices(
                dict(
                    filter(
                        lambda item: item[0].axis == axis,
                        slices.items(),
                    )
                )
            )

    def setVolumeShape(self, shape: tuple[int]) -> None:
        self._XYPlot2D.setSliceRange(0, shape[0] - 1)
        self._XZPlot2D.setSliceRange(0, shape[1] - 1)
        self._YZPlot2D.setSliceRange(0, shape[2] - 1)

    def getVolumeShape(self) -> tuple[int]:
        return (
            self._XYPlot2D.getSliceRange()[1] + 1,
            self._XZPlot2D.getSliceRange()[1] + 1,
            self._YZPlot2D.getSliceRange()[1] + 1,
        )

    def setVolumeMetadata(self, metadata: dict, volume_shape: tuple[int]):
        """
        Warning: as the call might change the scale it must be called once the data is set.
        """
        self._geometryOrMetadataWidget.setMetadata(metadata=metadata)
        self._updatePlotAxes(metadata=metadata, volume_shape=volume_shape)

    def _updatePlotAxes(self, metadata: dict, volume_shape: tuple[int]):
        """
        Define axes limits and labels according to the volume metadata (when possible)
        """
        reconstruction_metadata = metadata.get("processing_options", {}).get(
            "reconstruction", {}
        )
        position_m = reconstruction_metadata.get("position", None)
        voxel_size_cm = reconstruction_metadata.get("voxel_size_cm", None)

        labels = ("Z", "Y", "X")
        if position_m is not None and voxel_size_cm is not None:
            voxel_size = numpy.array(voxel_size_cm) * _ureg.centimeter
        else:
            _logger.info("Missing metadata from volume. Cannot set axis values")
            voxel_size = [None] * 3
        # Note: pixel size must be a scalar. So we arbitrarly choice one of the two values. Voxel are expected to be squares
        self._XYPlot2D.setVoxelSize(voxel_size[2])
        self._XZPlot2D.setVoxelSize(voxel_size[2])
        self._YZPlot2D.setVoxelSize(voxel_size[1])

        map_silx_axis_to_index_axis: {Axis, int} = {
            self._XYPlot2D.getPlot().getXAxis(): 1,
            self._XYPlot2D.getPlot().getYAxis(): 2,
            self._YZPlot2D.getPlot().getXAxis(): 1,
            self._YZPlot2D.getPlot().getYAxis(): 0,
            self._XZPlot2D.getPlot().getXAxis(): 2,
            self._XZPlot2D.getPlot().getYAxis(): 0,
        }
        for silx_axis, axis_index in map_silx_axis_to_index_axis.items():
            silx_axis.setLabel(labels[axis_index])

    def _popUpImageFullScreen(self, plot: TomoPlot2DSlices):
        new_plot = FullScreenPlot2D()
        active_image = plot.getPlot().getActiveImage()
        if active_image is None or active_image.getData(copy=False) is None:
            return

        window_title = f"Plot of {plot.getTitle()}"

        new_plot.setWindowTitle(window_title)

        # add the current image
        new_plot.addImage(
            active_image.getData(copy=False), colormap=active_image.getColormap()
        )

        new_plot.showFullScreen()

    def getMasterPlot(self) -> _TomoPlot2D:
        """
        To apply the same colormap to the three plot we need to have a master / driver one.
        It has been decided arbitrarily to use the
        """
        return self._YZPlot2D.getPlot()

    def resetZooms(self):
        for plot in self.getPlots().values():
            plot.resetZoom()

    def setVolumeIdentifier(self, volume_identifier: VolumeIdentifier):
        self._geometryOrMetadataWidget.setVolumeIdentifier(volume_identifier)



from tomwer.gui.utils.host_editor import HostEditor
from tomwer.tests.conftest import qtapp  # noqa F401


def test_HostEditor(
    qtapp,  # noqa F811
):
    # Note: host pinging is not compliant with CI
    widget = HostEditor(name=None, enable_host_pinging=False)
    widget.setCurrentText("tata")
    widget.setPort(26)

    assert widget.currentText() == "tata"



"""
test vignettes
"""

import time
import numpy
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.reconstruction.scores.scores import ComputedScore, ScoreMethod
from tomwer.gui.utils.vignettes import VignettesQDialog
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestVignettes(TestCaseQt):
    def setUp(self):
        super().setUp()
        self._window = VignettesQDialog(
            value_name="cor",
            score_name="score",
        )

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        super().tearDown()

    def test(self):
        self._window.setScores(
            scores={
                "3": (numpy.ones((100, 100)), ComputedScore(tv=2.0, std=3.2)),
                "7": (numpy.zeros((100, 100)), ComputedScore(tv=25, std=0.6)),
            },
            score_method="standard deviation",
        )
        self._window.show()
        self.qWaitForWindowExposed(self._window)

        def assert_scores_are_equal(scores_1, scores_2):
            assert scores_1.keys() == scores_2.keys()
            for key, (data_1, computed_score_1) in scores_1.items():
                data_2, computed_score_2 = scores_2[key]
                numpy.testing.assert_array_almost_equal(data_1, data_2)
                assert computed_score_1 == computed_score_2

        scores, score_method = self._window._vignettesWidget.getScores()
        assert_scores_are_equal(
            scores_1=scores,
            scores_2={
                "3": (numpy.ones((100, 100)), ComputedScore(std=3.2, tv=None)),
                "7": (numpy.zeros((100, 100)), ComputedScore(std=0.6, tv=None)),
            },
        )
        assert score_method is ScoreMethod.STD

        self._window.setNbColumn(4)
        self._window.setNbColumn(2)
        with pytest.raises(ValueError):
            self._window.setNbColumn(20)
        self._window.setNbColumn("auto")

        self._window.resize(qt.QSize(400, 400))
        self.qapp.processEvents()
        time.sleep(VignettesQDialog.RESIZE_MAX_TIME * 2 / 1000)
        self.qapp.processEvents()



# coding: utf-8
from __future__ import annotations


import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.saaxis import SAAxisTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.saaxis import QSAAxisParams

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class SAAxisProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisTask` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, saaxis_params, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)
        assert saaxis_params is not None
        self._dry_run = False
        self._process_fct = None

    def patch_processing(self, process_fct):
        self._computationThread.patch_processing(process_fct)

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(f"start saaxis {data}")
        assert isinstance(data, TomwerScanBase)
        if data.axis_params is None:
            data.axis_params = QAxisRP()
        if data.saaxis_params is None:
            data.saaxis_params = QSAAxisParams()
        self._data_currently_computed = data
        saaxis_params = QSAAxisParams.from_dict(configuration)
        saaxis_params.frame_width = data.dim_1
        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = f"low memory, skip saaxis calculation {data.path}"
            ProcessManager().notify_dataset_state(
                dataset=data, process=self._process_id, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            data.axis_params.set_relative_value(None)
            if callback is not None:
                callback()
            self.scan_ready(scan=data)
        else:
            self._data_currently_computed = data
            self._computationThread.init(data=data, saaxis_params=saaxis_params)
            # need to manage connect before starting it because
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.start()

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        assert isinstance(data, TomwerScanBase)
        FIFO._end_computation(
            self, data=data, future_tomo_obj=future_tomo_obj, callback=callback
        )

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of the axis position
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self.center_of_rotation = None
        self._dry_run = False
        self._scan = None
        self._saaxis_params = None
        self._patch_process_fct = None
        """function pointer to know which function to call for the axis
        calculation"""
        self._current_processing = None
        # processing currently runned. As we want to be able to cancel processing we need to keep a pointer on the
        # 'activate' nabu processing to stop process if requested
        self._task = None
        # ewoks task for reconstruction

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def patch_processing(self, process_fct):
        self._patch_process_fct = process_fct

    def init(self, data, saaxis_params: dict):
        self._scan = data
        self._saaxis_params = saaxis_params

    def run(self):
        self.sigComputationStarted.emit()
        if self._patch_process_fct:
            scores = {}
            for cor in self._saaxis_params.cors:
                scores[cor] = self._patch_process_fct(cor)
            self._scan.saaxis_params.scores = scores
            SAAxisTask.autofocus(scan=self._scan)
            self.center_of_rotation = self._scan.saaxis_params.autofocus
        else:
            self._task = SAAxisTask(
                process_id=self.process_id,
                inputs={
                    "dump_process": False,
                    "data": self._scan,
                    "dry_run": self._dry_run,
                    "sa_axis_params": self._saaxis_params.to_dict(),
                    "serialize_output_data": False,
                },
            )

            self._task.run()
            self.center_of_rotation = self._task.outputs.best_cor

    def cancel(self):
        if self._task is not None:
            self._task.cancel()
        self.quit()



from __future__ import annotations
import os
import logging
from tomwer.io.utils.raw_and_processed_data import (
    to_processed_data_path,
)
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.output import ProcessDataOutputDirMode
from tomwer.core.utils.scanutils import format_output_location
from tomwer.core.process.output import NabuOutputFileFormat

_logger = logging.getLogger(__name__)


PROCESS_FOLDER_RECONSTRUCTED_VOLUMES = "reconstructed_volumes"

PROCESS_FOLDER_RECONSTRUCTED_SLICES = "reconstructed_slices"

PROCESS_FOLDER_CAST_VOLUME = "cast_volume"


def get_output_folder_from_scan(
    mode: ProcessDataOutputDirMode,
    scan: TomwerScanBase,
    nabu_location: str | None,
    file_basename: str,
    file_format: NabuOutputFileFormat,
    processed_data_folder_name: str | None,
) -> tuple[str, str]:
    """

    :param mode: output mode, should save this to raw data, processed data...
    :param scan: scan for which we want to get the output folder
    :param nabu_location: output location provided by user (in case mode is 'other')
    :param file_basename: file basename to take in order to create output files
    :param file_format: output volume format (edf...)
    :param processed_data_folder_name: name of the processed data folder. Like 'reconstructed_volumes' or 'reconstructed_slices'...
    :return: (location, location_cfg_files). Location is the nabu configuration field 'output/location' 'location_cfg_files' is the information on where to save the nabu configuration file

    """
    output_mode = ProcessDataOutputDirMode.from_value(mode)
    file_format = NabuOutputFileFormat.from_value(file_format)

    if output_mode is ProcessDataOutputDirMode.OTHER and nabu_location in ("", None):
        _logger.error(
            "'other' output dir requested but no path provided. Fall back on the output dir to the scan folder"
        )
        # note: this is only an info because we expect to pass by this one for all .ows configuration (before 1.3 version)
        # as there was no different option by the time
        output_mode = ProcessDataOutputDirMode.IN_SCAN_FOLDER

    if output_mode is ProcessDataOutputDirMode.OTHER:
        location = format_output_location(nabu_location, scan=scan)
        location_cfg_files = location
    elif output_mode in (
        ProcessDataOutputDirMode.IN_SCAN_FOLDER,
        ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER,
    ):
        # otherwise default location will be the data root level
        location = os.path.join(scan.path, processed_data_folder_name)
        location_cfg_files = location
        if file_format in (
            NabuOutputFileFormat.EDF,
            NabuOutputFileFormat.TIFF,
            NabuOutputFileFormat.JP2K,
        ):  # if user specify the location
            location = "/".join([location, file_basename])
        if output_mode is ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER:
            location = to_processed_data_path(location)
            location_cfg_files = to_processed_data_path(location_cfg_files)
    else:
        raise NotImplementedError(f"mode {output_mode.value} is not implemented yet")

    return location, location_cfg_files



# coding: utf-8

import os
import tempfile
import unittest

from nxtomo.application.nxtomo import NXtomo
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockNXtomo


class TestMockNXtomo(unittest.TestCase):
    """Test the MockNXtomo file"""

    def test_creation(self):
        folder = tempfile.mkdtemp()
        mock = MockNXtomo(scan_path=folder, n_proj=10, n_ini_proj=10)
        self.assertEqual(
            mock.scan_master_file,
            os.path.join(folder, os.path.basename(folder) + ".h5"),
        )
        tomoScan = NXtomoScan(mock.scan_path, entry=mock.scan_entry)
        self.assertEqual(len(NXtomo.get_valid_entries(mock.scan_master_file)), 1)
        tomoScan.update()
        self.assertEqual(tomoScan.scan_range, 360)
        self.assertEqual(len(tomoScan.projections), 10)



import h5py
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from tomwer.core.process.reconstruction.axis import AxisTask
from tomwer.tests.conftest import nxtomo_scan_360  # noqa F811
from tomwer.core.process.reconstruction.axis import AxisRP


def test_read_x_rotation_axis_pixel_position(nxtomo_scan_360):  # noqa F811
    """
    test reading of the estimated cor from the motor using scan metadata
    """
    nexus_paths = get_nexus_paths(None)

    x_rotation_axis_pixel_position_path = "/".join(
        [
            nxtomo_scan_360.entry,
            nexus_paths.INSTRUMENT_PATH,
            nexus_paths.nx_instrument_paths.DETECTOR_PATH,
            nexus_paths.nx_detector_paths.X_ROTATION_AXIS_PIXEL_POSITION
            or nexus_paths.nx_detector_paths.ESTIMATED_COR_FRM_MOTOR_PATH,
        ]
    )
    with h5py.File(nxtomo_scan_360.master_file, mode="r") as h5f:
        assert x_rotation_axis_pixel_position_path not in h5f

    axis_params = AxisRP()
    axis_params.mode = "read"
    task = AxisTask(
        inputs={
            "data": nxtomo_scan_360,
            "axis_params": axis_params,
        }
    )
    # test the task when there is no metadata
    task.run()
    assert nxtomo_scan_360.axis_params.absolute_cor_value is None
    assert nxtomo_scan_360.axis_params.relative_cor_value is None

    # test the task when there is metadata
    with h5py.File(nxtomo_scan_360.master_file, mode="a") as h5f:
        h5f[x_rotation_axis_pixel_position_path] = 12.5

    nxtomo_scan_360.clear_cache()
    task.run()
    assert nxtomo_scan_360.axis_params.absolute_cor_value == 22.5
    assert nxtomo_scan_360.axis_params.relative_cor_value == 12.5



from __future__ import annotations

import numpy
import os
import shutil

from .dracbase import DracDatasetBase
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.esrf.scan.utils import from_bliss_original_file_to_raw
from tomwer.core.utils.dictutils import concatenate_dict
from tomwer.io.utils.raw_and_processed_data import to_raw_data_path

__all__ = [
    "DracRawDataset",
]


class DracRawDataset(DracDatasetBase):
    """
    Class to associate a raw dataset (TomwerScanBase) to an icat (raw) dataset
    """

    def __init__(
        self,
        tomo_obj: TomwerScanBase,
        raw_darks_required: bool = True,
        raw_flats_required: bool = True,
        raw_projections_required: bool = True,
        raw_projections_each: float = 90,
    ) -> None:
        if not isinstance(tomo_obj, TomwerScanBase):
            raise TypeError(
                f"tomo_obj should be an instance of {TomwerScanBase}. Got {type(tomo_obj)}"
            )
        data_dir = DracRawDataset.get_data_dir(scan=tomo_obj)
        super().__init__(data_dir=data_dir, tomo_obj=tomo_obj)
        self.__raw_darks_required = raw_darks_required
        self.__raw_flats_required = raw_flats_required
        self.__raw_projections_each = raw_projections_each
        self.__raw_projections_required = raw_projections_required

    @property
    def metadata(self) -> dict:
        scan = self.tomo_obj
        assert isinstance(scan, TomwerScanBase)
        if scan is not None:
            return scan.build_drac_metadata()
        else:
            return {}

    @property
    def bliss_raw_datasets(self) -> tuple[str]:
        return super().from_scan_to_raws(scan=self.tomo_obj)

    @property
    def raw_darks_required(self) -> bool:
        return self.__raw_darks_required

    @property
    def raw_flats_required(self) -> bool:
        return self.__raw_flats_required

    @property
    def raw_projections_each(self) -> float:
        return self.__raw_projections_each

    @property
    def raw_projections_required(self) -> bool:
        return self.__raw_projections_required

    def build_gallery(self):
        if os.path.exists(self.get_gallery_dir()):
            if self.gallery_overwrite:
                shutil.rmtree(self.get_gallery_dir())
            else:
                raise RuntimeError(
                    f"Gallery output dir ({self.get_gallery_dir()} already exists and cannot overwrite it.)"
                )
        scan = self.tomo_obj
        # collect all the data to be saved
        data_to_save: dict[str, numpy.ndarray] = {}
        if self.raw_darks_required and len(scan.darks) > 0:
            data_to_save["dark"] = next(iter(scan.darks.values()))
        if self.raw_flats_required and len(scan.flats) > 0:
            data_to_save["flat"] = next(iter(scan.flats.values()))
        if self.raw_projections_required:
            projections = scan.projections_with_angle()
            picked_angles = DracRawDataset.select_angles(
                angles_list=sorted(projections.keys()),
                each_angle=self.raw_projections_each,
            )
            data_to_save = concatenate_dict(
                data_to_save,
                {
                    f"projection_{angle:.1f}": projections[angle]
                    for angle in picked_angles
                },
            )

        # dump to disk
        for img_name, img_data in data_to_save.items():
            self.save_to_gallery(
                output_file_name=os.path.join(self.get_gallery_dir(), img_name),
                image=img_data,
            )

    @staticmethod
    def select_angles(angles_list: tuple, each_angle: int) -> tuple:
        angles_list = sorted(angles_list)
        if len(angles_list) > 0:
            start_angle = angles_list[0]
            stop_angle = angles_list[-1]
            picked_angle = numpy.arange(start_angle, stop_angle + 1, step=each_angle)
            return tuple(
                [
                    DracRawDataset.get_closest_projection(angle, angles_list)
                    for angle in picked_angle
                ]
            )
        else:
            return tuple()

    @staticmethod
    def get_closest_projection(angle, angles_list):
        idx_closest = numpy.argmin(numpy.abs(angles_list - angle))
        return angles_list[idx_closest]

    @staticmethod
    def get_data_dir(scan: HDF5TomoScan) -> str | None:
        """Deduce (bliss) data directory from a scan.
        If the scan can get the information from 'get_bliss_original_files' then we use it as this is most reliable information.
        Else we use the nxtomo location as a fallback (in case the data doesn't comes from bliss for example).
        """
        # the most reliable way is to get it from 'get_bliss_original_files' as this is set during conversion
        # else
        raw_data_file = scan.get_bliss_original_files()
        if scan.master_file is None:
            return None
        else:
            raw_data_file = to_raw_data_path(scan.master_file)
        return from_bliss_original_file_to_raw(os.path.dirname(raw_data_file))



"""
(Ewoks)Orange Widgets to 'control' the flow. Like defining input / output, stacking scans...
Those widget are not doing any 'core' processing.
"""

NAME = "control"

ID = "orangecontrib.tomwer.widgets.control"

DESCRIPTION = "widgets for data control"

LONG_DESCRIPTION = "Contains widget to control the scan flow"

ICON = "../../widgets/icons/tomwer_control.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 1



from __future__ import annotations
import logging

from tomwer.core.process.task import Task
from tomwer.core.process.drac.dracbase import DracDatasetBase

from processview.core.manager import DatasetState, ProcessManager

try:
    from pyicat_plus.client.main import IcatClient  # noqa F401
except ImportError:
    has_pyicat_plus = False
else:
    has_pyicat_plus = True

_logger = logging.getLogger(__name__)

__all__ = [
    "PublishICatDatasetTask",
]


class PublishICatDatasetTask(
    Task,
    input_names=(
        "data_portal_processed_datasets",
        "beamline",
        "proposal",
        "dataset",
    ),
    optional_input_names=(
        "dry_run",
        "__process__",
    ),
):
    """
    publish a list of 'IcatDataBase' instances.

    `IcatDataBase` provide API to retrieve data and metadata to be publish

    input field:
    * data_portal_processed_datasets: list of 'DracDatasetBase' instances.
    * beamline: name of the beamline (bm05, id19...)
    * proposal: proposal name
    """

    def run(self):
        for icat_data in self.inputs.data_portal_processed_datasets:
            if not isinstance(icat_data, DracDatasetBase):
                raise TypeError(f"icat_data should be an instance of {DracDatasetBase}")

            # build gallery (if needed by the icat data)

            process = self.get_input_value("__process__", None)

            if process is not None and icat_data.tomo_obj is not None:
                ProcessManager().notify_dataset_state(
                    dataset=icat_data.tomo_obj,
                    process=process(),
                    state=DatasetState.ON_GOING,
                )

            icat_data.build_gallery()

            try:
                self.publish_to_data_portal(
                    path=icat_data.data_dir,
                    metadata=icat_data.metadata,
                    raw=icat_data.bliss_raw_datasets,
                    dataset=icat_data.dataset_name,
                )

            except Exception as e:
                if process is not None and icat_data.tomo_obj is not None:
                    ProcessManager().notify_dataset_state(
                        dataset=icat_data.tomo_obj,
                        process=process(),
                        state=DatasetState.FAILED,
                    )
                    raise e
            else:
                if process is not None and icat_data.tomo_obj is not None:
                    ProcessManager().notify_dataset_state(
                        dataset=icat_data.tomo_obj,
                        process=process(),
                        state=DatasetState.SUCCEED,
                    )

    def publish_to_data_portal(self, path: str, metadata: dict, raw: str, dataset: str):
        """publish path to data_portal (drac) with given metadata"""
        if not self.inputs.dry_run:
            if not has_pyicat_plus:
                raise RuntimeError(
                    "pyicat_plus not installed it. Please install it to be able to publish dataset to icat"
                )
            icat_client = IcatClient(
                metadata_urls=("bcu-mq-01.esrf.fr:61613", "bcu-mq-02.esrf.fr:61613")
            )

            _logger.info(
                "publish to icat: %s",
                {
                    "path": path,
                    "beamline": self.inputs.beamline,
                    "proposal": self.inputs.proposal,
                    "raw": raw,
                    "dataset": dataset,
                    "metadata": metadata,
                },
            )
            icat_client.store_processed_data(
                beamline=self.inputs.beamline,
                proposal=self.inputs.proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
                raw=raw,
            )



from __future__ import annotations

import fnmatch
import logging
import os
import shutil

from tomoscan.esrf.scan.utils import get_unique_files_linked
import tomwer.version
from tomwer.core.utils.deprecation import deprecated_warning
from tomwer.core.process.reconstruction.nabu.settings import NABU_CFG_FILE_FOLDER
from tomwer.core.process.reconstruction.nabu.utils import update_cfg_file_after_transfer
from tomwer.core.process.task import Task
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.settings import get_dest_path, get_lbsram_path
from tomwer.core.signal import Signal
from tomwer.core.utils import logconfig
from tomwer.core.utils.spec import rebaseParFile
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.io.utils import get_linked_files_with_entry
from tomwer.core.process.reconstruction.saaxis.saaxis import (
    DEFAULT_RECONS_FOLDER as MULTI_COR_DEFAULT_FOLDER,
)
from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
    DEFAULT_RECONS_FOLDER as MULTI_DB_DEFAULT_FOLDER,
)

logger = logging.getLogger(__name__)

try:
    from tomwer.synctools.rsyncmanager import RSyncManager
except ImportError:
    logger.warning("rsyncmanager not available")
    has_rsync = False
else:
    has_rsync = True


class ScanTransferTask(
    Task,
    input_names=("data",),
    optional_input_names=(
        "serialize_output_data",
        "copying",
        "block",
        "turn_off_print",
        "dest_dir",
        "move",
        "noRsync",
        "overwrite",
    ),
    output_names=("data",),
):
    """Manage the copy of scan.

    .. warning : the destination directory is find out from the file system
                 if /lbsramxxx exists for example...
                 In the case we couldn't found the output directory then we
                 will ask for the user to set it.
    """

    scanready = Signal(TomwerScanBase)
    """emit when scan ready"""

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        super().__init__(
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        self.turn_off_print = inputs.get("turn_off_print", False)
        self._destDir = inputs.get("dest_dir", None)
        """
        output directory if forced. By default based by the env variable
        'TARGET_OUTPUT_FOLDER' if exists, else set to '/data/visitor'
         """
        self._copying = inputs.get("copying", False)
        force_sync = inputs.get("force_sync", None)
        if force_sync is not None:
            from tomwer.core.utils.deprecation import deprecated_warning

            deprecated_warning(
                type_="Parameter",
                name="force_sync",
                reason="Two parameters for the same option",
                replacement="block",
            )
        else:
            force_sync = False

        self._block = self.get_input_value("block", force_sync)

        self._move = self.get_input_value("move", False)
        if not isinstance(self._move, bool):
            raise TypeError("move is expected to be a boolean")

        self._overwrite = self.get_input_value("overwrite", False)
        if not isinstance(self._overwrite, bool):
            raise TypeError("move is expected to be a boolean")

        self._noRsync = self.get_input_value(
            "noRsync", False
        )  # TODO: rename noRsync to no_rsync
        if not isinstance(self._noRsync, bool):
            raise TypeError("move is expected to be a boolean")

    def set_configuration(self, properties):
        # No properties stored for now
        if "dest_dir" in properties:
            self.setDestDir(properties["dest_dir"])

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "data transfer"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "transfer data from a folder to another"

    @staticmethod
    def getDefaultOutputDir():
        """Return the default output dir based on the computer setup"""
        if "TARGET_OUTPUT_FOLDER" in os.environ:
            return os.environ["TARGET_OUTPUT_FOLDER"]

        if os.path.isdir("/data/visitor"):
            return "/data/visitor"

        return ""

    def _process_edf_scan(self, scan, move=False, overwrite=True, noRsync=False):
        if not isinstance(scan, EDFTomoScan):
            raise TypeError(f"{scan} is expected to be an instance of {EDFTomoScan}")
        outputdir = self.getDestinationDir(scan.path)
        if outputdir is None:
            return

        self._pretransfertOperations(scan.path, outputdir)
        # as we are in the workflow we want this function to be bloking.
        # so we will not used a thread for folder synchronization
        # for now rsync is not delaing with overwrite option
        if not has_rsync or noRsync is True or RSyncManager().has_rsync() is False:
            logger.info("Can't use rsync, copying files")
            try:
                if move is True:
                    self._moveFiles(
                        scanPath=scan.path,
                        outputdir=os.path.dirname(outputdir),
                        overwrite=overwrite,
                    )
                else:
                    self._copyFiles(
                        scanPath=scan.path, outputdir=outputdir, overwrite=overwrite
                    )
            except shutil.Error as e:
                raise e
            else:
                output_scan = scan._deduce_transfert_scan(outputdir)
                try:
                    output_scan._update_latest_recons_identifiers(
                        old_path=scan.path, new_path=output_scan.path
                    )
                except Exception as e:
                    logger.warning(
                        "Fail to convert url of latest reconstruction. Reason is:" + e
                    )
                    output_scan.clear_latest_reconstructions()
                self.__noticeTransfertSuccess(input_scan=scan, output_scan=output_scan)
        else:
            source = scan.path
            if not source.endswith(os.path.sep):
                source = source + os.path.sep
            target = outputdir

            if not target.endswith(os.path.sep):
                target = target + os.path.sep

            self._signalCopying(scanID=source, outputdir=target)
            output_scan = scan._deduce_transfert_scan(outputdir)
            try:
                output_scan._update_latest_recons_identifiers(
                    old_path=scan.path, new_path=output_scan.path
                )
            except Exception as e:
                logger.warning(
                    "Fail to convert url of latest reconstruction. Reason is:" + e
                )
                output_scan.clear_latest_reconstructions()
            RSyncManager().sync_file(
                source=source,
                target=target,
                wait=self._block,
                delete=True,
                callback=self.__noticeTransfertSuccess,
                callback_parameters=(scan, output_scan),
                rights=777,
            )
        return output_scan

    def _get_hdf5_dst_scan(self, bliss_scan_folder_path):
        if self._destDir is not None:
            rel_path = os.path.join(*bliss_scan_folder_path.split(os.sep)[-2:])
            return os.path.dirname(os.path.join(self._destDir, rel_path))
        # try to get outputdir from spec
        scanIDPath = os.path.abspath(bliss_scan_folder_path)
        return self._getOutputDirSpec() or os.path.join(
            *self._getOutputDirLBS(scanIDPath).split(os.sep)[:-1]
        )

    def _get_hdf5_sample_file_or_nx_dst(self, bliss_sample_file):
        bliss_sample_file = os.path.abspath(bliss_sample_file)
        if self._destDir is not None:
            rel_path = os.path.join(*bliss_sample_file.split(os.sep)[-2:])
            return os.path.join(self._destDir, rel_path)
        # try to get outputdir from spec
        bliss_sample_file = os.path.abspath(bliss_sample_file)
        return self._getOutputDirSpec() or self._getOutputDirLBS(bliss_sample_file)

    def _get_master_sample_file_dst(self, master_sample_file):
        if self._destDir is not None:
            rel_path = os.path.join(*master_sample_file.split(os.sep)[-2:])
            return os.path.join(self._destDir, rel_path)
        # try to get outputdir from spec
        master_sample_file = os.path.abspath(master_sample_file)
        return self._getOutputDirSpec() or self._getOutputDirLBS(master_sample_file)

    def _get_hdf5_proposal_file_dst(self, bliss_proposal_file):
        if self._destDir is not None:
            return os.path.join(self._destDir, os.path.split(bliss_proposal_file)[-1])
        # try to get outputdir from spec
        bliss_sample_file = os.path.abspath(bliss_proposal_file)
        return self._getOutputDirSpec() or self._getOutputDirLBS(bliss_sample_file)

    def _process_hdf5_scan(self, scan) -> TomwerScanBase:
        assert isinstance(scan, NXtomoScan)
        logger.warning(
            "scan transfer for HDF5 is a prototype. Please check transfer is properly executed."
        )

        files_sources = []
        files_dest = []
        delete_opt = []

        associated_files = get_unique_files_linked(url=scan.get_detector_url())
        associated_files = filter(
            lambda file_: os.path.abspath(file_) == os.path.abspath(scan.master_file),
            associated_files,
        )

        # manage .nx file
        if os.path.exists(scan.master_file):
            (
                new_nx_file,
                files_sources_for_nx,
                files_dest_for_nx,
            ) = self.handle_nexus_file(scan.master_file, scan.entry)
            files_sources.extend(files_sources_for_nx)
            files_dest.extend(files_dest_for_nx)
            delete_opt.append([True] * len(files_dest_for_nx))
            files_sources.append(scan.master_file)
            files_dest.append(new_nx_file)
            delete_opt.append(True)
            output_scan = NXtomoScan(scan=new_nx_file, entry=scan.entry)
        else:
            output_scan = None

        # manage files generated (*slice*.h5/edf, *.cfg, *.par...)
        # for reconstructed file, .h5, .edf if there is some conflict at one
        # point I guess we might need to check file entry ? or rename the file
        # according to the entry.

        # manage .par, .cfg and .rec files if any
        patterns = ["*.par", "*.cfg", "*.rec", "*.log"]
        # manage *slice*.hdf5 and *slice*.edf files (reconstructed slice)
        patterns += [
            "*slice*.hdf5",
            "*slice*.h5",
            "*slice*.jpeg",
            "*slice*.jpg",
            "*slice*.tiff",
            "*slice*.tif",
            "*slice*.j2k",
        ]
        # manage *vol files
        patterns += [
            "*_vol",
            "*_vol.hdf5",
        ]
        # manage nabu and tomwer processes files
        patterns += [
            "*tomwer_processes.h5",
            "*nabu_processes.hdf5",
            "steps_file_basename_nabu_sinogram_save_step.hdf5",
        ]
        # manage new dark and flat files
        patterns += [
            "*dark.hdf5",
            "*darks.hdf5",
            "*flats.hdf5",
        ]

        def match_pattern(file_name):
            file_name = file_name.lower()
            for pattern in patterns:
                if fnmatch.fnmatch(file_name, pattern):
                    return True
            return False

        dir_name = os.path.dirname(scan.master_file)
        for file_ in os.listdir(dir_name):
            if match_pattern(file_name=file_):
                full_file_path = os.path.join(dir_name, file_)
                files_sources.append(full_file_path)
                files_dest.append(self._get_hdf5_sample_file_or_nx_dst(full_file_path))
                delete_opt.append(True)

        # manage folders
        patterns = [
            NABU_CFG_FILE_FOLDER,
            "*slice*",
            MULTI_COR_DEFAULT_FOLDER,
            MULTI_DB_DEFAULT_FOLDER,
            "steps_file_basename_nabu_sinogram_save_step",
        ]
        for folder_ in os.listdir(dir_name):
            if match_pattern(file_name=folder_):
                full_file_path = os.path.join(dir_name, folder_)
                files_sources.append(full_file_path)
                # don't know why we need this os.sep ? call another get dest ?
                # os.path.join(os.sep, self._get_hdf5_dst_scan(full_file_path), folder_)
                dest = os.path.join(os.sep, self._get_hdf5_dst_scan(full_file_path))
                files_dest.append(dest)
                delete_opt.append(True)

        RSyncManager().sync_files(
            sources=files_sources,
            targets=files_dest,
            wait=self._block,
            delete=delete_opt,
            callback=self.__noticeTransfertSuccess,
            callback_parameters=(scan, output_scan),
        )
        return output_scan

    def handle_nexus_file(self, src_nexus_file: str, entry: str) -> tuple:
        """
        Return a tuple of three elemts:
        src_nexus_file, files_sources_for_nx, files_dest_for_nx

        :param src_nexus_file: source file to treat
        :return: (dst_nexus_file, files_sources_for_nx, files_dest_for_nx)
                * dst_nexus_file: str -> output nexus file that will be created by the transfer
                * list files_sources_for_nx: list of relative files connected to source nexus file
                * files_dest_for_nx: list of creating files from files_sources_for_nx once transfer is accomplish
        """
        dst_nx_file = self._get_hdf5_sample_file_or_nx_dst(src_nexus_file)
        linked_datasets_src = get_linked_files_with_entry(
            hdf5_file=src_nexus_file, entry=entry
        )
        linked_files_src = list(map(lambda x: x[0], linked_datasets_src))
        relative_linked_files_src = list(
            filter(lambda file_path: not os.path.isabs(file_path), linked_files_src)
        )
        relative_linked_files_dst = [
            os.path.normpath(os.path.join(os.path.dirname(dst_nx_file), file_path))
            for file_path in relative_linked_files_src
        ]
        relative_linked_files_src = [
            os.path.normpath(os.path.join(os.path.dirname(src_nexus_file), file_path))
            for file_path in relative_linked_files_src
        ]
        return dst_nx_file, relative_linked_files_src, relative_linked_files_dst

    def run(self):
        """Launch the process process

        :param scan: the path to the file we want to move/process
        :param move: if True, directly move the files. Otherwise copy the files
        :param overwrite: if True then force the copy even if the file/folder already
            exists
        :param noRSync: True if we wan't do sue shutil instead of rsync.
        """
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            raise ValueError("'scan' should be provided")

        if scan is None:
            self.outputs.data = None

        _scan = scan
        if type(_scan) is dict:
            _scan = ScanFactory.create_scan_object_frm_dict(scan)

        assert isinstance(_scan, TomwerScanBase)

        logger.info("synchronisation with scanPath")
        if isinstance(scan, EDFTomoScan):
            output_scan = self._process_edf_scan(
                scan=scan,
                move=self._move,
                overwrite=self.get_input_value("overwrite", True),
                noRsync=self._noRsync,
            )
        elif isinstance(scan, NXtomoScan):
            if self._move is True:
                raise NotImplementedError("move option not implemented")
            if self._noRsync is True:
                raise NotImplementedError("noRsync option not implemented")
            output_scan = self._process_hdf5_scan(scan=scan)
        else:
            raise TypeError("Other scan than EDF or HDF5 are not managed")
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = output_scan.to_dict()
        else:
            self.outputs.data = output_scan

    def _pretransfertOperations(self, scanfolder, outputdir):
        """Operation to be run before making the transfert of the scan"""
        self._updateParFiles(scanfolder, outputdir)
        self._updateNabuConfigFiles(scanfolder, outputdir)

    def _updateNabuConfigFiles(self, scanfolder, outputdir):
        nabu_cfg_folder = os.path.join(scanfolder, NABU_CFG_FILE_FOLDER)
        if os.path.exists(nabu_cfg_folder):
            for _file in os.listdir(nabu_cfg_folder):
                try:
                    update_cfg_file_after_transfer(
                        config_file_path=os.path.join(nabu_cfg_folder, _file),
                        old_path=scanfolder,
                        new_path=outputdir,
                    )
                except Exception:
                    pass

    def _updateParFiles(self, scanfolder, outputdir):
        """Update all path contained in the .par files to fit the new outpudir"""
        if not os.path.exists(scanfolder):
            return
        for _file in os.listdir(scanfolder):
            if _file.lower().endswith(".par"):
                rebaseParFile(
                    os.path.join(scanfolder, _file),
                    oldfolder=scanfolder,
                    newfolder=outputdir,
                )

    def __noticeTransfertSuccess(self, input_scan, output_scan):
        self._signalCopySucceed()

        logger.processSucceed(
            f"transfer succeed of {input_scan} to {output_scan}",
            extra={
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.FROM: str(input_scan),
                logconfig.TO: str(output_scan),
            },
        )
        self.signalTransfertOk(input_scan=input_scan, output_scan=output_scan)

    def signalTransfertOk(self, input_scan, output_scan):
        if input_scan is None or output_scan is None:
            return
        assert isinstance(input_scan, TomwerScanBase)
        assert isinstance(output_scan, TomwerScanBase)
        self.scanready.emit(output_scan)

    def _copyFiles(self, scanPath, outputdir, overwrite):
        """Copying files and removing them"""
        assert type(scanPath) is str
        assert type(outputdir) is str
        assert os.path.isdir(scanPath)
        # create the destination dir
        if not os.path.isdir(outputdir):

            def createDirAndTopDir(_dir):
                if not os.path.isdir(os.path.dirname(_dir)):
                    createDirAndTopDir(os.path.dirname(_dir))
                os.makedirs(_dir)

            createDirAndTopDir(outputdir)
        # we can't copy directly the top folder because he is already existing
        for f in os.listdir(scanPath):
            file = os.path.join(scanPath, f)
            fileDest = os.path.join(outputdir, f)
            if overwrite is True:
                if os.path.isdir(fileDest):
                    shutil.rmtree(fileDest)
                if os.path.isfile(fileDest):
                    os.remove(fileDest)
            if os.path.exists(fileDest):
                raise FileExistsError(fileDest, "already exists")
            if os.path.isdir(file):
                shutil.copytree(src=file, dst=fileDest)
            else:
                shutil.copy2(src=file, dst=fileDest)

        info = "Removing directory at %s" % scanPath
        logger.info(info)
        shutil.rmtree(scanPath)
        info = "sucessfuly removed file at %s !!!" % scanPath
        logger.info(info)

    def _moveFiles(self, scanPath, outputdir, overwrite):
        """Function simply moving files"""
        assert os.path.isdir(scanPath)

        logger.debug(
            "synchronisation with scanPath",
            extra={logconfig.DOC_TITLE: self._scheme_title},
        )

        target = os.path.join(outputdir, os.path.basename(scanPath))
        if overwrite is True and os.path.isdir(target):
            shutil.rmtree(target)
        shutil.move(scanPath, outputdir)

    def _requestFolder(self):
        out = None
        while out is None:
            out = input("please give the output directory : \n")
            if not os.path.isdir(out):
                warning = "given path " + out
                warning += " is not a directory, please give a valid directory"
                logger.warning(warning)
                out = None
        return out

    def _getOutputDirSpec(self):
        return None

    def _getOutputDirLBS(self, scanPath):
        if scanPath.startswith(get_lbsram_path()):
            return scanPath.replace(get_lbsram_path(), get_dest_path(), 1)
        else:
            return None

    def getDestinationDir(self, scanPath, ask_for_output=True):
        """Return the destination directory. The destination directory is the
        root directory"""
        if self._destDir is not None:
            return os.path.join(self._destDir, os.path.basename(scanPath))

        # try to get outputdir from spec
        scanIDPath = os.path.abspath(scanPath)

        outputdir = self._getOutputDirSpec() or self._getOutputDirLBS(scanIDPath)
        if outputdir is None and ask_for_output:
            outputdir = self._requestFolder()

        return outputdir

    def setDestDir(self, dist):
        """Force the outpudir to dist.

        :param dist: path to the folder. If None remove overwrite behavior
        """
        self._destDir = dist
        if self._destDir is not None and os.path.isdir(self._destDir):
            logger.warning("Given path %s is not a directory" % self._destDir)

    # some function to print the output in the terminal #

    def _signalCopying(self, scanID, outputdir):
        self._copying = True
        if self.turn_off_print is False:
            print("######################################")
            print("###")
            print("###")
            print("### copying files ", scanID, " to ", outputdir)
            print("### ...")

        info = f"start moving folder from {scanID} to {outputdir}"
        logger.processStarted(info, extra={logconfig.DOC_TITLE: self._scheme_title})

    def _signalCopyFailed(self):
        self._copying = False
        if self.turn_off_print is False:
            print("###")
            print("### copy failed")
            print("###")
            print("######################################")

    def _signalCopySucceed(self):
        self._copying = False
        if self.turn_off_print is False:
            print("###")
            print("### copy succeeded")
            print("###")
            print("######################################")

    def isCopying(self):
        """

        :return: True if the folder transfert is actually doing a copy
        """
        return self._copying

    def setForceSync(self, b):
        """

        :param b: if True then folderTransfert will wait until transfert
            is done to be released. Otherwise will launch a 'free' thread wich
            will notice transfert end later.
        """
        self._block = b

    def getOutput(self, scan):
        """

        :param scan:
        :return:
        """
        return os.path.join(self.getDestinationDir(scan), os.path.basename(scan))


class ScanTransfer(ScanTransferTask):
    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        deprecated_warning(
            name="tomwer.core.process.control.scantransfer.ScanTransfer",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="ScanTransferTask",
        )
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)



from __future__ import annotations

import argparse
import logging
import os
import shutil
import signal
import sys
import pkg_resources

import silx
from contextlib import closing
from logging.handlers import RotatingFileHandler
from xml.sax.saxutils import escape

import pyqtgraph
from orangecanvas.application.canvasmain import DockWidget
from orangecanvas.application.outputview import (
    ExceptHook,
    TerminalTextDocument,
    TextStream,
)
from orangecanvas.document.usagestatistics import UsageStatistics
from orangecanvas.main import Main as ocMain
from orangewidget.workflow import config
from orangewidget.workflow.config import data_dir_base
from orangewidget.workflow.errorreporting import handle_exception
from orangewidget.workflow.mainwindow import OWCanvasMainWindow as _MainWindow
from processview.gui.processmanager import ProcessManagerWindow
from silx.gui import qt

import tomwer.version
from orangecanvas.preview import previewdialog, previewmodel
from orangecanvas.application import examples as orange_examples
import tomoscan.version

from .config import TomwerConfig, TomwerSplashScreen

from tomwer.core.log.processlog import (
    PROCESS_ENDED_NAME,
    PROCESS_FAILED_NAME,
    PROCESS_INFORM_NAME,
    PROCESS_SKIPPED_NAME,
    PROCESS_STARTED_NAME,
    PROCESS_SUCCEED_NAME,
)

try:
    import nabu
except ImportError:
    has_nabu = False
else:
    has_nabu = True
try:
    import nxtomomill.version
except ImportError:
    has_nxtomomill = False
else:
    has_nxtomomill = True
try:
    import nxtomo
except ImportError:
    has_nxotmo = False
else:
    has_nxotmo = True
try:
    import sluurp
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True

_logger = logging.getLogger(__file__)

MAX_LOG_FILE = 10
"""Maximal log file kepts for orange"""

LOG_FILE_NAME = "tomwer.log"

LOG_FOLDER = "/var/log/tomwer"


# These are the sequences need to get colored ouput
_RESET_SEQ = "\033[0m"

_BLACK = "\033[30m"
_RED = "\033[31m"
_GREEN = "\033[32m"
_YELLOW = "\033[33m"
_BLUE = "\033[34m"
_MAGENTA = "\033[35m"


LOG_COLORS = {
    "WARNING": _MAGENTA,
    "INFO": _BLACK,
    "DEBUG": _BLUE,
    "CRITICAL": _YELLOW,
    "ERROR": _RED,
    PROCESS_SKIPPED_NAME: _MAGENTA,
    PROCESS_ENDED_NAME: _BLACK,
    PROCESS_INFORM_NAME: _BLACK,
    PROCESS_STARTED_NAME: _BLACK,
    PROCESS_FAILED_NAME: _RED,
    PROCESS_SUCCEED_NAME: _GREEN,
}


class MainWindow(_MainWindow):
    HELPDESK_URL = "https://requests.esrf.fr/plugins/servlet/desk/portal/41"

    FORCE_OBJECT_SUPERVISOR_DISPLAY = True
    """If true will always display the processed supervisor at start up. Restore state will be ignored."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.process_supervisor_dock = DockWidget(
            self.tr("object supervisor"),
            self,
            objectName="processes-dock",
            allowedAreas=qt.Qt.BottomDockWidgetArea,
            visible=self.show_processes_manager_action.isChecked(),
        )

        self.process_supervisor_dock.setWidget(ProcessManagerWindow(parent=None))
        self.process_supervisor_dock.visibilityChanged[bool].connect(
            self.show_processes_manager_action.setChecked
        )
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self.process_supervisor_dock)

    def restoreState(self, state, version=0):
        # for now we always want to display the process supervisor
        restored = super().restoreState(state=state, version=version)
        if self.FORCE_OBJECT_SUPERVISOR_DISPLAY:
            self.show_processes_manager_action.setChecked(True)
            self.process_supervisor_dock.setVisible(True)

        return restored

    def setup_actions(self):
        super().setup_actions()
        # create the action to connect with it
        self.show_processes_manager_action = qt.QAction(
            self.tr("&object supervisor"),
            self,
            toolTip=self.tr("Show object states relative to processes."),
            checkable=True,
            triggered=lambda checked: self.process_supervisor_dock.setVisible(checked),
        )

    def setup_menu(self):
        super().setup_menu()
        self.view_menu.addAction(self.show_processes_manager_action)

        self.simple_example_action = qt.QAction(
            self.tr("simple use cases"),
            self,
            objectName="simple-use-cases",
            toolTip=self.tr("Some simple examples."),
            triggered=self.open_simple_examples,
            menuRole=qt.QAction.AboutRole,
        )

        self.dark_flat_example_action = qt.QAction(
            self.tr("darks and flats manipulations"),
            self,
            objectName="dark-flat-manipulation",
            toolTip=self.tr("Examples around dark and flat."),
            triggered=self.open_dark_flat_examples,
            menuRole=qt.QAction.AboutRole,
        )
        self.slurm_execution_example_action = qt.QAction(
            self.tr("remote processing with slurm"),
            self,
            objectName="execution-on-slurm",
            toolTip=self.tr("Examples of some execution over slurm."),
            triggered=self.open_slurm_execution_example,
            menuRole=qt.QAction.AboutRole,
        )
        self.cor_search_example_action = qt.QAction(
            self.tr("cor search advance example"),
            self,
            objectName="cor-search-example",
            toolTip=self.tr("Examples of ways to search for the Center of rotation."),
            triggered=self.open_cor_search_example,
            menuRole=qt.QAction.AboutRole,
        )
        self.python_script_example_action = qt.QAction(
            self.tr("python script examples"),
            self,
            objectName="python-script-example",
            toolTip=self.tr("Examples of some python script usage."),
            triggered=self.open_python_script_example,
            menuRole=qt.QAction.AboutRole,
        )
        self.python_script_example_action = qt.QAction(
            self.tr("ID16b examples"),
            self,
            objectName="id16b-example",
            toolTip=self.tr("ID16b use case."),
            triggered=self.open_id16b_example,
            menuRole=qt.QAction.AboutRole,
        )
        # Predefined workflows menu.
        self.predefined_workflows_menu = qt.QMenu(
            self.tr("&Examples"),
            self.menuBar(),
            objectName="examples-menu",
        )
        self.predefined_workflows_menu.addActions(
            [
                self.simple_example_action,
                self.cor_search_example_action,
                self.dark_flat_example_action,
                self.slurm_execution_example_action,
                self.python_script_example_action,
            ]
        )

        self.menuBar().addMenu(self.predefined_workflows_menu)

        self._helpdeskButton = qt.QAction(
            self.tr("&Helpdesk"),
            self,
            objectName="helpdeskButton",
            triggered=self._contactUs,
            toolTip="contact esrf data processing helpdesk if you have some troubles with the application",
        )
        self.menuBar().addAction(self._helpdeskButton)

    def _contactUs(self, *args, **kwargs):
        try:
            qt.QDesktopServices.openUrl(qt.QUrl(self.HELPDESK_URL))
        except Exception as e:
            _logger.error(
                f"Failed to launch helpdesk web page ({self.HELPDESK_URL}). Error is {e}"
            )

    def __open_examples(self, examples, filter_names=None):
        """
        filter the existing tutorials by names contained in the file name
        """
        items = [previewmodel.PreviewItem(path=t.abspath()) for t in examples]

        def my_filter(value):
            for filter_name in filter_names:
                if filter_name in value:
                    return True
            return False

        if filter_names is not None:
            items = filter(
                lambda item: my_filter(os.path.basename(item.path())),
                items,
            )
        dialog = previewdialog.PreviewDialog(self)
        model = previewmodel.PreviewModel(dialog, items=items)
        title = self.tr("Example Workflows")
        dialog.setWindowTitle(title)
        template = '<h3 style="font-size: 26px">\n' "{0}\n" "</h3>"

        dialog.setHeading(template.format(title))
        dialog.setModel(model)

        model.delayedScanUpdate()
        status = dialog.exec()
        index = dialog.currentIndex()

        dialog.deleteLater()

        if status == qt.QDialog.Accepted:
            selected = model.item(index)
            self.open_example_scheme(selected.path())
        return status

    def open_about(self):
        dlg = AboutDialog(self)
        dlg.setAttribute(qt.Qt.WA_DeleteOnClose)
        dlg.exec()

    def open_simple_examples(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("simple", "EBS_tomo_listener")
        )

    def open_slurm_execution_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("slurm",)
        )

    def open_dark_flat_examples(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("darks",)
        )

    def open_python_script_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("_script",)
        )

    def open_id16b_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("ID16b",)
        )

    def open_cor_search_example(self):
        return self.__open_examples(
            orange_examples.workflows(), filter_names=("find_cor", "cor_search")
        )


log = logging.getLogger(__name__)


def check_for_updates() -> bool:
    return False


def send_usage_statistics() -> bool:
    return False


def pull_notifications() -> bool:
    return False


def make_sql_logger(level=logging.INFO):
    sql_log = logging.getLogger("sql_log")
    sql_log.setLevel(level)
    handler = RotatingFileHandler(
        os.path.join(config.log_dir(), "sql.log"), maxBytes=1e7, backupCount=2
    )
    sql_log.addHandler(handler)


class _OMain(ocMain):
    DefaultConfig = "tomwer.app.canvas_launcher.launcher.TomwerConfig"

    def run(self, argv):
        # Allow termination with CTRL + C
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        # Disable pyqtgraph's atexit and QApplication.aboutToQuit cleanup handlers.
        pyqtgraph.setConfigOption("exitCleanup", False)
        super().run(argv)

    def argument_parser(self) -> argparse.ArgumentParser:
        parser = super().argument_parser()
        parser.add_argument(
            "--use-opengl-plot",
            "--opengl-backend",
            help="Use OpenGL for plots (instead of matplotlib)",
            action="store_true",
            default=False,
        )
        return parser

    def parse_arguments(self, argv: list[str]):
        super().parse_arguments(argv)
        # define silx plot backend. For now the si;pler is to
        # define an environment variable.
        # this way we should also be able to let the user define it
        # on it's own later...

        if self.options.use_opengl_plot:

            # other standalones are not passing by the silx.config
            # but for the canvas this is way simpler to use it.
            # then all plot which have backend==None will pick the default plot backend
            silx.config.DEFAULT_PLOT_BACKEND = "gl"
        else:
            silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    def setup_logging(self):
        level = self.options.log_level
        logformat = "%(asctime)s:%(levelname)s:%(name)s: %(message)s"

        # File handler should always be at least INFO level so we need
        # the application root level to be at least at INFO.
        root_level = min(level, logging.INFO)
        rootlogger = logging.getLogger()
        rootlogger.setLevel(root_level)

        # Standard output stream handler at the requested level
        if sys.platform in ("linux", "linux2", "darwin"):
            stream_hander = make_linux_stream_colored_handler(
                level, fileobj=self.__stderr__
            )
        else:
            stream_hander = make_stream_handler(
                level, fileobj=self.__stderr__, fmt=logformat
            )
        rootlogger.addHandler(stream_hander)
        # Setup log capture for MainWindow/Log
        log_stream = TextStream(objectName="-log-stream")
        self.output.connectStream(log_stream)
        self.stack.push(closing(log_stream))  # close on exit
        log_handler = make_stream_handler(level, fileobj=log_stream, fmt=logformat)
        rootlogger.addHandler(log_handler)

        # Also log to file
        file_handler = make_file_handler(
            root_level,
            os.path.join(config.log_dir(), "canvas.log"),
            mode="w",
        )
        rootlogger.addHandler(file_handler)

        make_sql_logger(self.options.log_level)

    def setup_application(self):
        super().setup_application()
        # NOTE: No OWWidgetBase subclass should be imported before this

        self._update_check = check_for_updates()
        self._send_stat = send_usage_statistics()
        self._pull_notifs = pull_notifications()

        settings = qt.QSettings()
        settings.setValue(
            "startup/launch-count",
            settings.value("startup/launch-count", 0, int) + 1,
        )

        UsageStatistics.set_enabled(False)

    def show_splash_message(self, message: str, color=qt.QColor("#FFD39F")):
        super().show_splash_message(message, color)

    def create_main_window(self):
        window = MainWindow()
        return window

    def setup_sys_redirections(self):
        super().setup_sys_redirections()
        if isinstance(sys.excepthook, ExceptHook):
            sys.excepthook.handledException.connect(handle_exception)

    def tear_down_sys_redirections(self):
        if isinstance(sys.excepthook, ExceptHook):
            sys.excepthook.handledException.disconnect(handle_exception)
        super().tear_down_sys_redirections()

    def splash_screen(self):
        """Return the application splash screen"""
        settings = qt.QSettings()
        options = self.options
        want_splash = (
            settings.value("startup/show-splash-screen", True, type=bool)
            and not options.no_splash
        )

        if want_splash:
            pm, rect = self.config.splash_screen()
            splash_screen = TomwerSplashScreen(pixmap=pm, textRect=rect)
            splash_screen.setAttribute(qt.Qt.WA_DeleteOnClose)
            splash_screen.setFont(qt.QFont("Helvetica", 12))
            palette = splash_screen.palette()
            color = qt.QColor("#b3baba")
            palette.setColor(qt.QPalette.Text, color)
            splash_screen.setPalette(palette)
        else:
            splash_screen = None
        return splash_screen


def data_dir():
    return os.path.join(data_dir_base(), "tomwer", tomwer.version.version)


def widget_settings_dir():
    return os.path.join(data_dir(), "widgets")


ABOUT_TEMPLATE = """\
<h4>{name}</h4>
<p>tomwer version: {tomwer_version}</p>
<p>nabu version: {nabu_version}</p>
<p>nxtomo version: {nxtomo_version}</p>
<p>nxtomomill version: {nxtomomill_version}</p>
<p>tomoscan version: {tomoscan_version}</p>
<p>ewokscore version: {ewokscore_version}</p>
<p>ewoksorange version: {ewoksorange_version}</p>
<p>sluurp version: {sluurp_version}</p>
"""


class AboutDialog(qt.QDialog):
    def __init__(self, parent=None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        layout = qt.QVBoxLayout()
        label = qt.QLabel(self)

        pixmap, _ = TomwerConfig.splash_screen()
        pixmap = pixmap.scaled(150, 150)

        label.setPixmap(pixmap)

        layout.addWidget(label, qt.Qt.AlignCenter)

        text = ABOUT_TEMPLATE.format(
            name=escape("tomwer"),
            tomwer_version=escape(tomwer.version.version),
            nabu_version=escape(nabu.version if has_nabu else "not installed"),
            nxtomomill_version=escape(
                nxtomomill.version.version if has_nxtomomill else "not installed"
            ),
            nxtomo_version=escape(
                nxtomo.__version__ if has_nxotmo else "not installed"
            ),
            tomoscan_version=escape(tomoscan.version.version),
            sluurp_version=escape(
                sluurp.__version__ if has_sluurp else "not installed"
            ),
            ewokscore_version=escape(
                pkg_resources.get_distribution("ewokscore").version
            ),
            ewoksorange_version=escape(
                pkg_resources.get_distribution("ewoksorange").version
            ),
        )
        text_label = qt.QLabel(text)
        layout.addWidget(text_label, qt.Qt.AlignCenter)

        buttons = qt.QDialogButtonBox(qt.QDialogButtonBox.Close, qt.Qt.Horizontal, self)
        layout.addWidget(buttons)
        buttons.rejected.connect(self.accept)
        layout.setSizeConstraint(qt.QVBoxLayout.SetFixedSize)
        self.setLayout(layout)


class OMain(_OMain):
    config: TomwerConfig
    DefaultConfig = "tomwer.app.canvas_launcher.config.TomwerConfig"

    def run(self, argv):
        log.info("Clearing widget settings")
        shutil.rmtree(widget_settings_dir(), ignore_errors=True)
        dealWithLogFile()
        super().run(argv)

    def setup_application(self):
        qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
        return super().setup_application()

    def setup_logging(self):
        super().setup_logging()
        # rootlogger = logging.getLogger()
        # rootlogger = TomwerLogger(rootlogger)
        # logging.setLoggerClass(TomwerLogger)

    def setup_sys_redirections(self):
        self.output = doc = TerminalTextDocument()

        stdout = TextStream(objectName="-stdout")
        stderr = TextStream(objectName="-stderr")
        doc.connectStream(stdout)
        doc.connectStream(stderr, color=qt.Qt.red)

        if sys.stdout is not None:
            stdout.stream.connect(sys.stdout.write, qt.Qt.DirectConnection)

        self.__stdout__ = sys.stdout
        sys.stdout = stdout

        if sys.stderr is not None:
            stderr.stream.connect(sys.stderr.write, qt.Qt.DirectConnection)

        self.__stderr__ = sys.stderr
        sys.stderr = stderr
        self.__excepthook__ = sys.excepthook
        sys.excepthook = ExceptHook(stream=stderr)

        self.stack.push(closing(stdout))
        self.stack.push(closing(stderr))

    def argument_parser(self) -> argparse.ArgumentParser:
        parser = super().argument_parser()
        for action in parser._actions:
            # avoid clearing settings because this make people confused
            if action.dest == "clear_widget_settings":
                parser._remove_action(action)
                break
            # by default log are only displaying errors. But we want to get warning as well (for deprecation for example).
            if action.dest == "log_level":
                action.default = logging.WARNING

        parser.add_argument(
            "--no-color-stdout-logs",
            "--no-colored-logs",
            action="store_true",
            help="instead of having logs in the log view, color logs of the stdout",
            default=False,
        )
        return parser

    def create_main_window(self):
        window = MainWindow()
        return window


def dealWithLogFile():
    """Move log file history across log file hierarchy and create the new log file"""

    # move log file if exists
    for i in range(MAX_LOG_FILE):
        logFile = LOG_FILE_NAME
        if os.path.exists(LOG_FOLDER) and os.access(LOG_FOLDER, os.W_OK):
            logFile = os.path.join(LOG_FOLDER, logFile)
        defLogName = logFile

        iLog = MAX_LOG_FILE - i
        maxLogNameN1 = logFile + "." + str(iLog)
        if iLog - 1 == 0:
            maxLogNameN2 = defLogName
        else:
            maxLogNameN2 = logFile + "." + str(iLog - 1)
        if os.path.exists(maxLogNameN2):
            try:
                stat = os.stat(maxLogNameN2)
                shutil.copy(maxLogNameN2, maxLogNameN1)
                os.utime(maxLogNameN1, (stat.st_atime, stat.st_mtime))
            except Exception:
                pass
    # create a new log file
    if os.path.exists(LOG_FOLDER) and os.access(LOG_FOLDER, os.W_OK):
        logFile = os.path.join(LOG_FOLDER, logFile)
        logging.basicConfig(
            filename=logFile,
            filemode="w",
            level=logging.WARNING,
            format="%(asctime)s %(message)s",
        )


class _ColoredFormatter(logging.Formatter):
    """Dedicated colors to highlight level name by coloring logs (only for linux terminals)"""

    @staticmethod
    def color_formatting(message: str, levelname: str, ascii_time):
        """colors message using ascii"""
        return (
            f"{ascii_time} {LOG_COLORS[levelname]}[{levelname}]{_RESET_SEQ} {message}"
        )

    def format(self, record):
        record.asctime = self.formatTime(record, "%Y-%m-%d %H:%M:%S")
        record.message = record.getMessage()
        return self.color_formatting(
            message=record.message,
            levelname=record.levelname,
            ascii_time=record.asctime,
        )


def make_linux_stream_colored_handler(level, fileobj=None):
    handler = logging.StreamHandler(fileobj)
    handler.setLevel(level)
    formatter = _ColoredFormatter()
    handler.setFormatter(formatter)
    return handler


def make_stream_handler(level, fileobj=None, fmt=None):
    handler = logging.StreamHandler(fileobj)
    handler.setLevel(level)
    if fmt:
        handler.setFormatter(logging.Formatter(fmt))
    return handler


def make_file_handler(level, filename, mode="w", fmt=None):
    handler = logging.FileHandler(filename, mode=mode, encoding="utf-8")
    handler.setLevel(level)
    if fmt:
        handler.setFormatter(logging.Formatter(fmt))
    return handler
